from dataclasses import dataclass, field
from typing import Dict
import numpy as np
if __package__ == "":
    from lux.cargo import UnitCargo
    from lux.config import EnvConfig
    from lux.team import Team, FactionTypes
    from lux.unit import Unit
    from lux.factory import Factory
else:
    from .cargo import UnitCargo
    from .config import EnvConfig
    from .team import Team, FactionTypes
    from .unit import Unit
    from .factory import Factory
def process_action(action):
    return to_json(action)
def to_json(state):
    if isinstance(state, np.ndarray):
        return state.tolist()
    elif isinstance(state, np.int64):
        return state.tolist()
    elif isinstance(state, list):
        return [to_json(s) for s in state]
    elif isinstance(state, dict):
        out = {}
        for k in state:
            out[k] = to_json(state[k])
        return out
    else:
        return state  
def from_json(state):
    if isinstance(state, list):
        return np.array(state)
    elif isinstance(state, dict):
        out = {}
        for k in state:
            out[k] = from_json(state[k])
        return out
    else:
        return state 

def process_obs(player, game_state, step, obs):
    if step == 0:
        # at step 0 we get the entire map information
        game_state = from_json(obs)
    else:
        # use delta changes to board to update game state
        obs = from_json(obs)
        for k in obs:
            if k != 'board':
                game_state[k] = obs[k]
        for item in ["rubble", "lichen", "lichen_strains"]:
            for k, v in obs["board"][item].items():
                k = k.split(",")
                x, y = int(k[0]), int(k[1])
                game_state["board"][item][y, x] = v
    return game_state

def obs_to_game_state(step, env_cfg: EnvConfig, obs):
    
    units = dict()
    for agent in obs["units"]:
        units[agent] = dict()
        for unit_id in obs["units"][agent]:
            unit_data = obs["units"][agent][unit_id]
            cargo = UnitCargo(**unit_data["cargo"])
            del unit_data["cargo"]
            units[agent][unit_id] = Unit(
                **unit_data,
                cargo=cargo,
                unit_cfg=env_cfg.ROBOTS[unit_data["unit_type"]],
                env_cfg=env_cfg
            )

    factory_occupancy_map = np.ones_like(obs["board"]["rubble"], dtype=int) * -1
    factories = dict()
    for agent in obs["factories"]:
        factories[agent] = dict()
        for unit_id in obs["factories"][agent]:
            f_data = obs["factories"][agent][unit_id]
            cargo = UnitCargo(**f_data["cargo"])
            del f_data["cargo"]
            factory = Factory(
                **f_data,
                cargo=cargo,
                env_cfg=env_cfg
            )
            factories[agent][unit_id] = factory
            factory_occupancy_map[factory.pos[1] - 1:factory.pos[1] + 1, factory.pos[0] - 1:factory.pos[0] + 1] = factory.team_id
    teams = dict()
    for agent in obs["teams"]:
        team_data = obs["teams"][agent]
        faction = FactionTypes[team_data["faction"]]
        del team_data["faction"]
        teams[agent] = Team(**team_data, faction=faction, agent=agent)

    return GameState(
        env_cfg=env_cfg,
        env_steps=step,
        board=Board(
            rubble=obs["board"]["rubble"],
            ice=obs["board"]["ice"],
            ore=obs["board"]["ore"],
            lichen=obs["board"]["lichen"],
            lichen_strains=obs["board"]["lichen_strains"],
            factory_occupancy_map=factory_occupancy_map,
            factories_per_team=obs["board"]["factories_per_team"]
        ),
        weather_schedule=obs["weather_schedule"],
        units=units,
        factories=factories,
        teams=teams

    )

@dataclass
class Board:
    rubble: np.ndarray
    ice: np.ndarray
    ore: np.ndarray
    lichen: np.ndarray
    lichen_strains: np.ndarray
    factory_occupancy_map: np.ndarray
    factories_per_team: int
@dataclass
class GameState:
    """
    A GameState object at step env_steps. Copied from luxai2022/state/state.py
    """
    env_steps: int
    env_cfg: dict
    board: Board
    weather_schedule: np.ndarray = None
    units: Dict[str, Dict[str, Unit]] = field(default_factory=dict)
    factories: Dict[str, Dict[str, Factory]] = field(default_factory=dict)
    teams: Dict[str, Team] = field(default_factory=dict)
    @property
    def real_env_steps(self):
        """
        the actual env step in the environment, which subtracts the time spent bidding and placing factories
        """
        if self.env_cfg.BIDDING_SYSTEM:
            # + 1 for extra factory placement and + 1 for bidding step
            return self.env_steps - (self.board.factories_per_team + 1 + 1)
        else:
            return self.env_steps


    # various utility functions
    def is_day(self):
        return self.real_env_steps % self.env_cfg.CYCLE_LENGTH < self.env_cfg.DAY_LENGTH


import dataclasses
import json
import math
import os
import random
import sys
from os import path
from .agents import all_agents
from sys import path as syspath
from os import path as osp

# next two lines enables importing local packages e.g. luxai2022
__dir__ = osp.dirname(__file__)
syspath.append(__dir__)


import vec_noise

from luxai2022.env import LuxAI2022
import numpy as np

import copy
import json
def to_json(state):
    if isinstance(state, np.ndarray):
        return state.tolist()
    elif isinstance(state, np.int64):
        return state.tolist()
    elif isinstance(state, list):
        return [to_json(s) for s in state]
    elif isinstance(state, dict):
        out = {}
        for k in state:
            out[k] = to_json(state[k])
        return out
    else:
        return state
prev_step = 0
luxenv: LuxAI2022 = LuxAI2022(verbose=0, validate_action_space=False)
prev_obs = None
state_obs = None
def cleanup_dimensions():
    global dimension_process
    if dimension_process is not None:
        dimension_process.kill()
def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()
def interpreter(state, env):
    global luxenv, prev_obs, state_obs
    player_0 = state[0]
    player_1 = state[1]
    # filter out actions such as debug annotations so they aren't saved
    # filter_actions(state, env)
    ### 1.2: Initialize a blank state game if new episode is starting ###
    if env.done:
        if "seed" in env.configuration:
            seed = int(env.configuration["seed"])
        else:
            seed = math.floor(random.random() * 1e9);
            env.configuration["seed"] = seed
        if "episodeSteps" in env.configuration:
            episodeSteps = int(env.configuration["episodeSteps"])
        else:
            episodeSteps = 1001

        parsed_env_config = copy.deepcopy(env.configuration)
        parsed_env_config["max_episode_length"] = episodeSteps
        delete_keys = ["seed", "episodeSteps", "actTimeout", "runTimeout", "env_cfg"]
        
        for k in delete_keys:
            if k in parsed_env_config: del parsed_env_config[k]
        luxenv = LuxAI2022(validate_action_space=True, **parsed_env_config)
        _ = luxenv.reset(seed=seed)
        state_obs = luxenv.state.get_compressed_obs()

        env_cfg_json = dataclasses.asdict(luxenv.env_cfg)
        # del env_cfg_json["WEATHER_ID_TO_NAME"]
        env.configuration.env_cfg = env_cfg_json
        
        player_0.observation.player = "player_0"
        player_1.observation.player = "player_1"
        # TODO add observation optimizations here later
        player_0.observation.obs = json.dumps(to_json(state_obs))
        
        player_0.observation.width = luxenv.state.board.width
        player_0.observation.height = luxenv.state.board.height
        return state

    new_state_obs, rewards, dones, infos = luxenv.step({
        "player_0": player_0.action,
        "player_1": player_1.action
    })

    player_0.observation.player = "player_0"
    player_1.observation.player = "player_1"

    player_0.observation.obs = json.dumps(to_json(luxenv.state.get_change_obs(state_obs)))
    state_obs = new_state_obs["player_0"]
    
    player_0.observation.width = luxenv.state.board.width
    player_0.observation.height = luxenv.state.board.height

    player_0.reward = int(rewards["player_0"])
    player_1.reward = int(rewards["player_1"])
    player_0.observation.reward = int(player_0.reward)
    player_1.observation.reward = int(player_1.reward)
    if np.all([dones[k] for k in dones]):
        if player_0.status == "ACTIVE":
            player_0.status = "DONE"
        if player_1.status == "ACTIVE":
            player_1.status = "DONE"
    return state

def renderer(state, env):
    raise NotImplementedError("To render the replay, please set the render mode to json or html")


dir_path = path.dirname(__file__)
json_path = path.abspath(path.join(dir_path, "lux_ai_2022.json"))
with open(json_path) as json_file:
    specification = json.load(json_file)


def html_renderer():
    html_path = path.abspath(path.join(dir_path, "index.html"))
    return ("html_path", html_path)

agents = all_agents


import json
import math
import random
import sys
from os import path
from .agents import agents as all_agents
from subprocess import Popen, PIPE
import atexit
from .test_agents.python.lux.game import Game
from threading import Thread
from queue import Queue, Empty

t = None
q = None
dimension_process = None
game_state = Game()
prev_step = 0
def cleanup_dimensions():
    global dimension_process
    if dimension_process is not None:
        dimension_process.kill()
def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()
def interpreter(state, env):
    global dimension_process, game_state, t, q, prev_step
    player1 = state[0]
    player2 = state[1]

    ### 1.1: Initialize dimensions in the background within the orchestrator if we haven't already ###
    if dimension_process is None:
        # dimension_process = Popen(["ts-node", "-P", path.abspath(path.join(dir_path, "dimensions/tsconfig.json")), path.abspath(path.join(dir_path, "dimensions/run.ts"))], stdin=PIPE, stdout=PIPE)
        try:
            dimension_process = Popen(["node", path.abspath(path.join(dir_path, "dimensions/main.js"))], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        except FileNotFoundError:
            import warnings
            warnings.warn("Node not installed")
            return state

        # following 4 lines from https://stackoverflow.com/questions/375427/a-non-blocking-read-on-a-subprocess-pipe-in-python
        q = Queue()
        t = Thread(target=enqueue_output, args=(dimension_process.stdout, q))
        t.daemon = True # thread dies with the program
        t.start()
        atexit.register(cleanup_dimensions)

    # filter out actions such as debug annotations so they aren't saved
    filter_actions(state, env)
    
    ### 1.2: Initialize a blank state game if new episode is starting ###
    if env.done:
        # TODO: allow resetting to a specific state
        # print("Initialize game", "steps", len(env.steps), "prev_step", prev_step)
        # last_state = None
        # if prev_step >= len(env.steps):
        #     last_state = env.steps[-1]
        # prev_step = len(env.steps)
        # print("prev_step now", prev_step)
        if "seed" in env.configuration:
            seed = env.configuration["seed"]
        else:
            seed = math.floor(random.random() * 1e9);
            env.configuration["seed"] = seed
        if "loglevel" in env.configuration:
            loglevel = env.configuration["loglevel"]
        else:
            loglevel = 0 # warnings, 1: errors, 0: none
            env.configuration["loglevel"] = loglevel
        if "annotations" in env.configuration:
            annotations = env.configuration["annotations"]
        else:
            annotations = False # warnings, 1: errors, 0: none
            env.configuration["annotations"] = annotations
        
        if "width" in env.configuration:
            width = env.configuration["width"]
        else:
            width = -1 # -1 for randomly selected
            env.configuration["width"] = width
        if "height" in env.configuration:
            height = env.configuration["height"]
        else:
            height = -1 # -1 for randomly selected
            env.configuration["height"] = height
        
        initiate = {
            "type": "start",
            "agent_names": [], # unsure if this is provided?
            "config": env.configuration
        }
        # if last_state is not None:
        #     initiate["state"] = last_state
        dimension_process.stdin.write((json.dumps(initiate) + "\n").encode())
        dimension_process.stdin.flush()

        agent1res = get_message(dimension_process)
        agent2res = get_message(dimension_process)
        match_obs_meta = get_message(dimension_process)
       
        player1.observation.player = 0
        player2.observation.player = 1
        player1.observation.updates = agent1res
        
        # player2.observation.updates = agent2res # duplicated and not added
        player1.observation.globalCityIDCount = match_obs_meta["globalCityIDCount"]
        player1.observation.globalUnitIDCount = match_obs_meta["globalUnitIDCount"]
        player1.observation.width = match_obs_meta["width"]
        player1.observation.height = match_obs_meta["height"]

        game_state = Game()
        game_state._initialize(agent1res)

        return state
    # print("prev_step", prev_step, "stored steps", len(env.steps))
    # prev_step += 1
    
    ### 2. : Pass in actions (json representation along with id of who made that action), agent information (id, status) to dimensions via stdin
    dimension_process.stdin.write((json.dumps(state) + "\n").encode())
    dimension_process.stdin.flush()


    ### 3.1 : Receive and parse the observations returned by dimensions via stdout
    agent1res = json.loads(dimension_process.stderr.readline())
    agent2res = json.loads(dimension_process.stderr.readline())
    game_state._update(agent1res)

    # receive meta info such as global ID and map sizes for purposes of being able to start from specific state
    match_obs_meta = json.loads(dimension_process.stderr.readline())
    match_status = json.loads(dimension_process.stderr.readline())

    while True:
        try:  line = q.get_nowait()
        except Empty:
            # no standard error received, break
            break
        else:
            # standard error output received, print it out
            print(line.decode(), file=sys.stderr, end='')

    ### 3.2 : Send observations to each agent through here. Like dimensions, first observation can include initialization stuff, then we do the looping

    player1.observation.updates = agent1res

    player1.observation.globalCityIDCount = match_obs_meta["globalCityIDCount"]
    player1.observation.globalUnitIDCount = match_obs_meta["globalUnitIDCount"]
    player1.observation.width = match_obs_meta["width"]
    player1.observation.height = match_obs_meta["height"]
    # player2.observation.updates = agent2res # duplicated and not added

    player1.observation.player = 0
    player2.observation.player = 1

    ### 3.3 : handle rewards
    # reward here is defined as the sum of number of city tiles
    player1.reward = compute_reward(game_state.players[0])
    player2.reward = compute_reward(game_state.players[1])
    player1.observation.reward = int(player1.reward)
    player2.observation.reward = int(player2.reward)

    ### 3.4 Handle finished match status
    if match_status["status"] == "finished":
        if player1.status == "ACTIVE":
            player1.status = "DONE"
        if player2.status == "ACTIVE":
            player2.status = "DONE"
    return state

def get_message(dimension_process):
    raw = dimension_process.stderr.readline()
    try:
        res = json.loads(raw)
        return res
    except Exception as e:
        print("Engine Exception")
        err_stack = dimension_process.stderr.readlines(100)
        # err_stack = [raw, *err_stack]
        # print(err_stack)
        for m in err_stack:
            if len(m) < 1000: 
                print(m.decode(), file=sys.stderr)
            else:
                print("...", file=sys.stderr)

def filter_actions(state, env):
    enable_annotations = env.configuration["annotations"]
    if not enable_annotations:
        for team in range(len(state)):
            filtered = []
            if state[team] is not None and state[team].action is not None:
                for l in state[team].action:
                    if len(l) > 0 and l[0] != "d":
                        filtered.append(l)
                state[team].action = filtered
        

def compute_reward(player):
    ct_count = sum([len(v.citytiles) for k, v in player.cities.items()])
    unit_count = len(game_state.players[player.team].units)
    # max board size is 32 x 32 => 1024 max city tiles and units, so this should keep it strictly so we break by city tiles then unit count
    return ct_count * 10000 + unit_count

def renderer(state, env):
    raise NotImplementedError("To render the replay, please set the render mode to json or html")


dir_path = path.dirname(__file__)
json_path = path.abspath(path.join(dir_path, "lux_ai_2021.json"))
with open(json_path) as json_file:
    specification = json.load(json_file)


def html_renderer():
    html_path = path.abspath(path.join(dir_path, "index.html"))
    return ("html_path", html_path)

agents = all_agents



import math
import sys
from typing import List
import numpy as np
from dataclasses import dataclass
if __package__ == "":
    from lux.weather import get_weather_config
    from lux.cargo import UnitCargo
    from lux.config import EnvConfig
else:
    from .weather import get_weather_config
    from .cargo import UnitCargo
    from .config import EnvConfig
# a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
move_deltas = np.array([[0, 0], [0, -1], [1, 0], [0, 1], [-1, 0]])

@dataclass
class Unit:
    team_id: int
    unit_id: str
    unit_type: str # "LIGHT" or "HEAVY"
    pos: np.ndarray
    power: int
    cargo: UnitCargo
    env_cfg: EnvConfig
    unit_cfg: dict
    action_queue: List

    def action_queue_cost(self, game_state):
        cost = self.env_cfg.UNIT_ACTION_QUEUE_POWER_COST[self.unit_type]
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        return cost * weather_cfg["power_loss_factor"]

    def move_cost(self, game_state, direction):
        board = game_state.board
        target_pos = self.pos + move_deltas[direction]
        if target_pos[0] < 0 or target_pos[1] < 0 or target_pos[1] >= len(board.rubble) or target_pos[0] >= len(board.rubble[0]):
            # print("Warning, tried to get move cost for going off the map", file=sys.stderr)
            return None
        factory_there = board.factory_occupancy_map[target_pos[1], target_pos[0]]
        if factory_there != self.team_id and factory_there != -1:
            # print("Warning, tried to get move cost for going onto a opposition factory", file=sys.stderr)
            return None
        rubble_at_target = board.rubble[target_pos[1]][target_pos[0]]
        
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        return math.ceil((self.unit_cfg.MOVE_COST + self.unit_cfg.RUBBLE_MOVEMENT_COST * rubble_at_target) * weather_cfg["power_loss_factor"])
    def move(self, direction, repeat=True):
        if isinstance(direction, int):
            direction = direction
        else:
            pass
        return np.array([0, direction, 0, 0, 1 if repeat else 0])

    def transfer(self, transfer_direction, transfer_resource, transfer_amount, repeat=True):
        assert transfer_resource < 5 and transfer_resource >= 0
        assert transfer_direction < 5 and transfer_direction >= 0
        return np.array([1, transfer_direction, transfer_resource, transfer_amount, 1 if repeat else 0])
    
    def pickup(self, pickup_resource, pickup_amount, repeat=True):
        assert pickup_resource < 5 and pickup_resource >= 0
        return np.array([2, 0, pickup_resource, pickup_amount, 1 if repeat else 0])
    
    def dig_cost(self, game_state):
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        return math.ceil(self.unit_cfg.DIG_COST * weather_cfg["power_loss_factor"])
    def dig(self, repeat=True):
        return np.array([3, 0, 0, 0, 1 if repeat else 0])

    def self_destruct_cost(self, game_state):
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        return math.ceil(self.unit_cfg.SELF_DESTRUCT_COST * weather_cfg["power_loss_factor"])
    def self_destruct(self, repeat=True):
        return np.array([4, 0, 0, 0, 1 if repeat else 0])

    def recharge(self, x, repeat=True):
        return np.array([5, 0, 0, x, 1 if repeat else 0])


# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from kaggle_environments import make, evaluate, utils

env = None


def custom1(obs):
    step = sum(1 for mark in obs.board if mark == obs.mark)
    return [0, 2, 4, 6, 8][step]


def custom2(obs):
    step = sum(1 for mark in obs.board if mark == obs.mark)
    return [1, 3, 5, 7][step]


def custom3(obs):
    step = sum(1 for mark in obs.board if mark == obs.mark)
    time.sleep(4)
    return [1, 3, 5, 7][step]


def custom4():
    raise Exception("Foo Bar")


def custom5():
    return -1


def custom6(obs):
    step = sum(1 for mark in obs.board if mark == obs.mark)
    time.sleep(2)
    return [1, 3, 5, 7][step]


def before_each(state=None):
    global env
    steps = [] if state == None else [state]
    env = make("tictactoe", steps=steps, debug=True)


def test_to_json():
    before_each()
    json = env.toJSON()
    assert json["name"] == "tictactoe"
    assert json["rewards"] == [0, 0]
    assert json["statuses"] == ["ACTIVE", "INACTIVE"]
    assert json["specification"]["reward"]["type"] == ["number", "null"]


def test_can_reset():
    before_each()
    assert env.reset() == [
        {
            "action": 0,
            "status": "ACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 1, "board": [0, 0, 0, 0, 0, 0, 0, 0, 0], "step": 0},
            "reward": 0,
        },
        {
            "action": 0,
            "status": "INACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_place_valid_mark():
    before_each()

    assert env.step([4, None]) == [
        {
            "action": 4,
            "status": "INACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 1, "board": [0, 0, 0, 0, 1, 0, 0, 0, 0], "step": 1},
            "reward": 0,
        },
        {
            "action": 0,  # None caused the default action to be applied.
            "status": "ACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_place_invalid_mark():
    before_each()

    env.step([4, None])

    assert env.step([None, 4]) == [
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 1, "board": [0, 0, 0, 0, 1, 0, 0, 0, 0], "step": 2},
            "reward": 0,
        },
        {
            "action": 4,
            "status": "INVALID",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "reward": None,
        },
    ]


def test_can_place_winning_mark():
    state1 = {"observation": {"board": [2, 1, 0, 1, 1, 0, 2, 0, 2]}}
    state2 = {}
    before_each([state1, state2])

    assert env.step([7, None]) == [
        {
            "action": 7,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 1, "board": [2, 1, 0, 1, 1, 0, 2, 1, 2], "step": 1},
            "reward": 1,
        },
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "reward": -1,
        },
    ]


def test_can_render():
    obs = {"observation": {"board": [0, 1, 0, 2, 1, 2, 0, 0, 2]}}
    before_each([obs, obs])
    out = "   | X |   \n---+---+---\n O | X | O \n---+---+---\n   |   | O "
    assert env.render(mode="ansi") == out


def test_can_step_through_agents():
    before_each()
    while not env.done:
        action1 = env.agents.random(env.state[0].observation)
        action2 = env.agents.reaction(
            utils.structify({"board": env.state[0].observation.board, "mark": 2}))
        env.step([action1, action2])
    assert env.state[0].reward + env.state[1].reward == 0


def test_can_run_agents():
    before_each()
    state = env.run(["random", "reaction"])[-1]
    assert state[0].reward + state[1].reward == 0


def test_can_evaluate():
    rewards = evaluate("tictactoe", ["random", "reaction"], num_episodes=2)
    assert (rewards[0][0] + rewards[0][1] ==
            0) and rewards[1][0] + rewards[1][1] == 0


def test_can_run_custom_agents():
    before_each()
    state = env.run([custom1, custom2])[-1]
    assert state == [
        {
            "action": 6,
            "reward": 1,
            "info": {},
            "observation": {"remainingOverageTime": 2, "board": [1, 2, 1, 2, 1, 2, 1, 0, 0], "mark": 1, "step": 7},
            "status": "DONE",
        },
        {
            "action": 0,
            "reward": -1,
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "status": "DONE",
        },
    ]


def test_agents_can_timeout_on_init():
    env = make("tictactoe", debug=True)
    state = env.run([custom1, custom3])[-1]
    assert state[1]["status"] == "TIMEOUT"
    assert state[1]["observation"]["remainingOverageTime"] < 0


def test_agents_can_timeout_on_act():
    env = make("tictactoe", debug=True)
    state = env.run([custom1, custom6])[-1]
    print(state)
    assert state[1]["status"] == "TIMEOUT"
    assert state[1]["observation"]["remainingOverageTime"] < 0


def test_run_timeout():
    env = make("tictactoe", debug=True, configuration={"actTimeout": 10, "runTimeout": 1})
    state = env.run([custom1, custom3])[-1]
    assert state == [
        {
            "action": 0,
            "reward": 0,
            "info": {},
            "observation": {"remainingOverageTime": 2, "board": [1, 2, 0, 0, 0, 0, 0, 0, 0], "mark": 1, "step": 2},
            "status": "ACTIVE",
        },
        {
            "action": 1,
            "reward": 0,
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "status": "INACTIVE",
        },
    ]


def test_agents_can_error():
    before_each()
    state = env.run([custom1, custom4])[-1]
    assert state == [
        {
            "action": 0,
            "reward": 0,
            "info": {},
            "observation": {"remainingOverageTime": 2, "board": [1, 0, 0, 0, 0, 0, 0, 0, 0], "mark": 1, "step": 2},
            "status": "DONE",
        },
        {
            "action": None,
            "reward": None,
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "status": "ERROR",
        },
    ]


def test_agents_can_have_invalid_actions():
    before_each()
    state = env.run([custom1, custom5])[-1]
    assert state == [
        {
            "action": 0,
            "reward": 0,
            "info": {},
            "observation": {"remainingOverageTime": 2, "board": [1, 0, 0, 0, 0, 0, 0, 0, 0], "mark": 1, "step": 2},
            "status": "DONE",
        },
        {
            "action": None,
            "reward": None,
            "info": {},
            "observation": {"remainingOverageTime": 2, "mark": 2},
            "status": "INVALID",
        },
    ]



import numpy as np
from scipy.ndimage import convolve, maximum_filter
from scipy.fft import dctn, idctn
from luxai2022.map_generator.visualize import viz
from luxai2022.map_generator.symnoise import SymmetricNoise, symmetrize

class GameMap(object):
    def __init__(self, rubble, ice, ore, symmetry):
        self.rubble = rubble
        self.ice = ice
        self.ore = ore
        self.symmetry = symmetry
        self.width, self.height = len(rubble[0]), len(rubble)

    @staticmethod
    def noise(seed=None, noise=None, noise_shift=None, symmetry=None):
        if seed is not None and noise is not None:
            raise ValueError("At most one of seed, noise can be specified.")

        if noise is None:
            noise = SymmetricNoise(seed=seed, symmetry=symmetry, octaves=3)
        else:
            seed = noise.seed
        if noise_shift is not None:
            noise.noise_shift += noise_shift
        if seed is not None and noise_shift is not None:
            noise.random.seed(seed + noise_shift)
        return noise

    @staticmethod
    def random_map(seed=None, map_type=None, symmetry=None, width=None, height=None):
        """random_map
        Returns a random map. Can specify seed.
        """
        noise = GameMap.noise(seed)
        if width is None:
            if height is None:
                width = noise.random.randint(32, 64)
            else:
                width = height
        if height is None:
            height = width
        if not map_type:
            map_type = noise.random.choice(["Cave", "Craters", "Island", "Mountain"])
        if not symmetry:
            symmetry = noise.random.choice(["horizontal", "vertical", "rotational", "/", "\\"])
        noise.update_symmetry(symmetry)
        return eval(map_type)(width, height, symmetry, noise=noise)

    @staticmethod
    def random_maps(n, seed=None):
        noise = GameMap.noise(seed)
        for i in range(n):
            noise = GameMap.noise(noise, noise_shift=100)
            width = height = noise.random.randint(32, 64)
            map_type = noise.random.choice(["Cave", "Craters", "Island", "Mountain"])
            symmetry = noise.random.choice(["horizontal", "vertical", "rotational", "/", "\\"])
            noise.update_symmetry(symmetry)

            yield eval(map_type)(width, height, symmetry, noise=noise)

class Cave(GameMap):
    def __init__(self, width=64, height=64, symmetry="vertical", seed=None, noise=None, noise_shift=None):
        """Cave
        Builds a cave system of size width x height and symmetrical across the `symmetry` axis.

        seed - If noise is not provided, use this seed to generate it.
        noise - A symmetric noise generator.
        noise_shift - Allows use of the same noise function to get different maps.

        Note: The last two are provided as it takes 3x longer to initialize SymmetricNoise than to run.
        """

        noise = GameMap.noise(seed, noise, noise_shift, symmetry)

        """
        Mask will end up with
           0 = interior
           1 = cave wall
           >1 = outside cave
        """
        # Start with mostly zeros
        mask = noise.random.randint(6, size=(height, width))
        mask[mask >= 1] = 1
        symmetrize(mask, symmetry)
        
        # Build clumps of ones (will be interior of caves)
        for i in range(3):
            mask = convolve(mask, [[1]*3]*3, mode="constant", cval=0) // 6
        mask = 1 - mask

        # Create cave wall
        mask += maximum_filter(mask, size=4)
        # fix bug where filter will cause it to be not symmetric...
        symmetrize(mask, symmetry)

        # Make some noisy rubble
        x = np.linspace(0, 1, width)
        y = np.linspace(0, 1, height)
        rubble = noise(x, y) * 50 + 50
        rubble = rubble.round()
        rubble[mask==1] //= 5 # Cave walls
        rubble[mask==0] = 0 # Interior of cave

        # Make some noisy ice, most ice is on cave edges
        ice = noise(x, y + 100)
        ice[mask > 1] = 0
        ice[mask == 0] = 0
        mid_mask = (ice < np.percentile(ice, 92)) & (ice > np.percentile(ice, 91))
        ice[ice < np.percentile(ice, 98)] = 0
        ice[ice != 0] = 1
        ice[mid_mask] = 1

        # Make some noisy ore, most ore is outside caves
        ore = noise(x, y - 100)
        ore[mask==1] = 0
        ore[mask==0] = 0
        mid_mask = (ore < np.percentile(ore, 92)) & (ore > np.percentile(ore, 91))
        ore[(ore < np.percentile(ore, 98))] = 0
        ore[ore != 0] = 1
        ore[mid_mask] = 1
        super().__init__(rubble, ice, ore, symmetry)

class Craters(GameMap):
    def __init__(self, width=64, height=64, symmetry="vertical", seed=None, noise=None, noise_shift=None):
        """Craters
        Builds a craters system of size `width` x `height` and symmetrical across the `symmetry` axis.

        seed - If noise is not provided, use this seed to generate it.
        noise - A symmetric noise generator.
        noise_shift - Allows use of the same noise function to get different maps.

        Note: The last two are provided as it takes 3x longer to initialize SymmetricNoise than to run.
        """

        noise = GameMap.noise(seed, noise, noise_shift, symmetry)

        min_craters = max(2, width*height // 1000)
        max_craters = max(4, width*height // 500)
        num_craters = noise.random.randint(min_craters, max_craters+1)

        # Mask = how many craters have hit the spot. When it symmetrizes, it will divide by 2.
        mask = np.zeros((height, width))
        x = np.linspace(0, 1, width)
        y = np.linspace(0, 1, height)
        crater_noise = noise(x, y, frequency=10) * 0.5 + 0.75 # Don't want perfectly circular craters.
        xx, yy = np.mgrid[:width, :height]

        # ice should be around edges of crater
        ice_mask = np.zeros((height, width))
        for i in range(num_craters):
            cx, cy = noise.random.randint(width), np.random.randint(height)
            cr = noise.random.randint(3, min(width, height)//4)
            c = (xx - cx) **2 + (yy - cy)**2
            c = c.T * crater_noise
            mask[c < cr**2] += 1
            edge = np.logical_and(c >= cr**2, c < 2 * cr**2)
            ice_mask[edge] = 2

        mask.astype(float)

        symmetrize(mask, symmetry)
        symmetrize(ice_mask, symmetry)

        rubble = (np.minimum(mask, 1) * 90 + 10) * noise(x, y+100, frequency=3)
        ice = noise(x, y-100)
        ice[ice_mask==0] = 0
        ice[ice < np.percentile(ice, 95)] = 0
        ice = np.round(50 * ice + 50)

        ore = np.minimum(mask, 1) * noise(x, y+100, frequency=3)
        ore[ore < np.percentile(ore, 95)] = 0
        ore = np.round(50 * ore + 50)

        super().__init__(rubble, ice, ore, symmetry)

def solve_poisson(f):
    """
    Solves the Poisson equation
        ∇²p = f
    using the finite difference method with Neumann boundary conditions.
    References: https://elonen.iki.fi/code/misc-notes/neumann-cosine/
                https://scicomp.stackexchange.com/questions/12913/poisson-equation-with-neumann-boundary-conditions
    """
    nx, ny = f.shape

    # Transform to DCT space
    dct = dctn(f, type=1)

    # Divide by magical factors
    cx = np.cos(np.pi * np.arange(nx) / (nx - 1))
    cy = np.cos(np.pi * np.arange(ny) / (ny - 1))
    f = np.add.outer(cx, cy) - 2

    np.divide(dct, f, out=dct, where=f != 0)
    dct[f == 0] = 0

    # Return to normal space
    potential = idctn(dct, type=1)
    return potential / 2

def nabla(x):
    return convolve(x, ((0, 1, 0), (1, -5, 1), (0, 1, 0)))

def dxx(x):
    return convolve(x, ((1,), (-2,), (1,)))

def dyy(x):
    return convolve(x, ((1, -2, 1),))

def dxy(x):
    return convolve(x, ((1, 0, -1), (0, 0, 0), (-1, 0, 1))) / 4

def laplacian(x):
    return convolve(x, ((1, 1, 1), (1, -8, 1), (1, 1, 1))) / 3



class Mountain(GameMap):
    def __init__(self, width=64, height=64, symmetry="vertical", seed=None, noise=None, noise_shift=None):
        """Mountain
        Builds a mountain system of size `width` x `height` and symmetrical across the `symmetry` axis.

        seed - If noise is not provided, use this seed to generate it.
        noise - A symmetric noise generator.
        noise_shift - Allows use of the same noise function to get different maps.

        Note: The last two are provided as it takes 3x longer to initialize SymmetricNoise than to run.
        """

        noise = GameMap.noise(seed, noise, noise_shift, symmetry)

        f = np.zeros((height, width))

        # Sprinkle a few mountains on the map.
        min_mountains = max(2, width*height//750)
        max_mountains = max(4, width*height//375)
        num_mountains = noise.random.randint(min_mountains, max_mountains+1)
        for i in range(num_mountains):
            x, y = noise.random.randint(width), noise.random.randint(height)
            f[y][x] -= 1

        symmetrize(f, symmetry)

        # mask will be floats in [0, 1], where 0 = no mountain, 1 = tallest peak
        mask = solve_poisson(f)
        symmetrize(mask, symmetry) # in case of floating point errors
        x = np.linspace(0, 1, width)
        y = np.linspace(0, 1, height)
        mask *= 5 + noise(x, y, frequency=3)
        mask -= np.amin(mask)

        # Find the valleys
        Lap = convolve(mask, ((1, 1, 1), (1, -8, 1), (1, 1, 1))) / 3 # Laplacian
        Dxx = convolve(mask, ((1,), (-2,), (1,)))
        Dyy = convolve(mask, ((1, -2, 1),))
        Dxy = convolve(mask, ((1, 0, -1), (0, 0, 0), (-1, 0, 1)))
        det = 16 * Dxx * Dyy - Dxy ** 2 # Hessian determinant
        det = det.astype(np.csingle) # complex

        cond = Lap * (2 * Lap - np.sqrt(4 * Lap**2 - det)) / det - 0.25 # ratio of eigenvalues
        cond = abs(cond) # should already be real except for floating point errors
        cond = np.maximum(cond, 1/cond)
        symmetrize(cond, symmetry) # for floating point errors

        def bdry(x, y):
            # Maybe instead of > 20, use some mean and std stuff?
            return abs(cond[y][x]) > 20 and f[y][x] == 0

        closed_set = set()
        def flood_fill(x, y):
            nonlocal closed_set
            region = []
            bounds = []
            open_set = {(x, y)}
            while len(open_set) > 0:
                new_set = set()
                closed_set |= open_set
                for x, y in open_set:
                    region.append((x, y))
                    for dx, dy in ((-1, 0), (1, 0), (0, -1), (0, 1)):
                        new_x, new_y = x + dx, y + dy
                        if (new_x, new_y) in closed_set:
                            continue
                        if 0 <= new_x < width and 0 <= new_y < height:
                            if bdry(new_x, new_y):
                                bounds.append((new_x, new_y))
                            else:
                                new_set.add((new_x, new_y))
                open_set = new_set
            return region, bounds

        regions = []
        bdrys = []
        for x in range(width):
            for y in range(height):
                if (x, y) in closed_set:
                    continue
                if bdry(x, y):
                    continue
                region, bounds = flood_fill(x, y)
                if len(region) > 0 and len(bounds) > 0:
                    regions.append(region)
                    bdrys.append(bounds)

        for region, bounds in zip(regions, bdrys):
            if len(region) * 5 > width * height:
                continue
            if np.mean(np.take(mask, region)) < 2 * np.mean(np.take(mask, bounds)):
                for x, y in region:
                    mask[y][x] = 0

        for x in range(width):
            for y in range(height):
                if bdry(x, y): mask[y][x] = 0

        mask[np.where(mask > 0)] -= np.amin(mask[np.where(mask>0)])
        mask -= np.amin(mask)
        mask /= np.amax(mask)

        rubble = (100*mask).round()
        ice = (100 * mask).round()
        mid_mask = (ice < np.percentile(ice, 50)) & (ice > np.percentile(ice, 48)) | (ice < np.percentile(ice, 20)) & (ice > np.percentile(ice, 0))
        ice[ice < np.percentile(ice, 99)] = 0
        ice[ice != 0] = 1
        ice[mid_mask] = 1
        
        ore = (100 * mask).round()
        mid_mask = (ore < np.percentile(ore, 62)) & (ore > np.percentile(ore, 60))
        ore[ore < np.percentile(ore, 83.5)] = 0
        ore[ore > np.percentile(ore, 84)] = 0
        ore[ore != 0] = 1
        ore[mid_mask] = 1

        super().__init__(rubble, ice, ore, symmetry)

class Island(GameMap):
    def __init__(self, width=64, height=64, symmetry="vertical", seed=None, noise=None, noise_shift=None):
        """Island
        Builds an island system of size `width` x `height` and symmetrical across the `symmetry` axis.

        seed - If noise is not provided, use this seed to generate it.
        noise - A symmetric noise generator.
        noise_shift - Allows use of the same noise function to get different maps.

        Note: The last two are provided as it takes 3x longer to initialize SymmetricNoise than to run.
        """

        noise = GameMap.noise(seed, noise, noise_shift, symmetry)

        # at the end, 0 = island, 1 = sea (of rubble)
        mask = noise.random.randint(4, size=(height, width))
        mask[mask >= 1] = 1

        s = -1
        while np.sum(mask==0) != s:
            s = np.sum(mask==0)
            mask = convolve(mask, [[1]*3]*3, mode="constant", cval=1) // 6

        # Shift every spot on the map a little, making the islands nosier.
        x = np.linspace(0, 1, width)
        y = np.linspace(0, 1, height)
        noise_dx = noise(x, y, frequency=10) * 4 - 2
        noise_dy = noise(x, y + 100, frequency=10) * 4 - 2

        xx, yy = np.mgrid[:width, :height]
        new_xx = xx + np.round(noise_dx).astype(int)
        new_xx = np.maximum(0, np.minimum(width-1, new_xx))
        new_yy = yy + np.round(noise_dy).astype(int)
        new_yy = np.maximum(0, np.minimum(height-1, new_yy))

        mask[yy, xx] = mask[new_yy, new_xx]

        symmetrize(mask, symmetry)

        rubble = noise(x, y - 100, frequency=3) * 50 + 50
        rubble[mask==0] //= 20

        # Unsure what to do about ice, ore right now. Place in pockets on islands?
        ice = noise(x, y + 200, frequency=10) ** 2 * 100
        ice[ice < 50] = 0
        ice[mask!=0] = 0

        ore = noise(x, y - 200, frequency=10) ** 2 * 100
        ore[ore < 50] = 0
        ore[mask!=0] = 0

        super().__init__(rubble, ice, ore, symmetry)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Generate maps for Lux AI 2022.")
    parser.add_argument("-t", "--map_type", help="Map type ('Cave', 'Craters', 'Island', 'Mountain')", required=False)
    parser.add_argument("-s", "--size", help="Size (32-64)", required=False)
    parser.add_argument("-d", "--seed", help="Seed")
    parser.add_argument("-m", "--symmetry", help="Symmetry ('horizontal', 'rotational', 'vertical', '/', '\\')")

    args = vars(parser.parse_args())
    map_type = args.get("map_type", None)
    symmetry = args.get("symmetry", None)
    if args.get("size", None):
        width = height = int(args["size"])
    else:
        width = height = None
    if args.get("seed", None):
        seed = int(args["seed"])
    else:
        seed = None
    game_map = GameMap.random_map(seed=seed, symmetry=symmetry, map_type=map_type, width=width, height=height)
    viz(game_map)
    input()



from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
import math
from typing import Callable

import numpy as np
from typing import TYPE_CHECKING
from luxai2022.config import EnvConfig
if TYPE_CHECKING:
    from luxai2022.factory import Factory
    from luxai2022.state import State
from luxai2022.map.position import Position


import luxai2022.unit as luxai_unit


# (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X, 6 = repeat)
class Action:
    def __init__(self, act_type: str) -> None:
        self.act_type = act_type
        self.repeat = False

    def state_dict():
        raise NotImplementedError("")


class FactoryBuildAction(Action):
    def __init__(self, unit_type: luxai_unit.UnitType) -> None:
        super().__init__("factory_build")
        self.unit_type = unit_type

    def state_dict(self):
        if self.unit_type == luxai_unit.UnitType.LIGHT:
            return 0
        return 1


class FactoryWaterAction(Action):
    def __init__(self) -> None:
        super().__init__("factory_water")
        self.water_cost = None

    def state_dict():
        return 2


class MoveAction(Action):
    def __init__(self, move_dir: int, dist: int = 1, repeat=False) -> None:
        super().__init__("move")
        # a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
        self.move_dir = move_dir
        self.dist = dist
        self.repeat = repeat

    def state_dict(self):
        return np.array([0, self.move_dir, self.dist, 0, self.repeat])


class TransferAction(Action):
    def __init__(self, transfer_dir: int, resource: int, transfer_amount: int, repeat=False) -> None:
        super().__init__("transfer")
        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)
        self.transfer_dir = transfer_dir
        self.resource = resource
        self.transfer_amount = transfer_amount
        self.repeat = repeat

    def state_dict(self):
        return np.array([1, self.transfer_dir, self.resource, self.transfer_amount, self.repeat])


class PickupAction(Action):
    def __init__(self, resource: int, pickup_amount: int, repeat=False) -> None:
        super().__init__("pickup")
        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)
        self.resource = resource
        self.pickup_amount = pickup_amount
        self.repeat = repeat

    def state_dict(self):
        return np.array([2, 0, self.resource, self.pickup_amount, self.repeat])


class DigAction(Action):
    def __init__(self, repeat=False) -> None:
        super().__init__("dig")
        self.repeat = repeat

    def state_dict(self):
        return np.array([3, 0, 0, 0, self.repeat])


class SelfDestructAction(Action):
    def __init__(self, repeat=False) -> None:
        super().__init__("self_destruct")
        self.repeat = repeat

    def state_dict(self):
        return np.array([4, 0, 0, 0, self.repeat])


class RechargeAction(Action):
    def __init__(self, power: int, repeat=False) -> None:
        super().__init__("recharge")
        self.power = power
        self.repeat = repeat

    def state_dict(self):
        return np.array([5, 0, 0, self.power, self.repeat])


def format_factory_action(a: int):
    if a == 0 or a == 1:
        unit_type = luxai_unit.UnitType.HEAVY
        if a == 0:
            unit_type = luxai_unit.UnitType.LIGHT
        return FactoryBuildAction(unit_type=unit_type)
    elif a == 2:
        return FactoryWaterAction()
    else:
        raise ValueError(f"Action {a} for factory is invalid")


def format_action_vec(a: np.ndarray):
    # (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X, 6 = repeat)
    a_type = a[0]
    if a_type == 0:
        return MoveAction(a[1], dist=1, repeat=a[4])
    elif a_type == 1:
        return TransferAction(a[1], a[2], a[3], repeat=a[4])
    elif a_type == 2:
        return PickupAction(a[2], a[3], repeat=a[4])
    elif a_type == 3:
        return DigAction(repeat=a[4])
    elif a_type == 4:
        return SelfDestructAction(repeat=a[4])
    elif a_type == 5:
        return RechargeAction(a[3], repeat=a[4])
    else:
        raise ValueError(f"Action {a} is invalid type, {a[0]} is not valid")


# a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
move_deltas = np.array([[0, 0], [0, -1], [1, 0], [0, 1], [-1, 0]])


def validate_actions(env_cfg: EnvConfig, state: 'State', actions_by_type, weather_cfg, verbose=1):
    """
    validates actions and logs warnings for any invalid actions. Invalid actions are subsequently not evaluated
    """
    actions_by_type_validated = defaultdict(list)
    valid_action = True

    def invalidate_action(msg):
        nonlocal valid_action
        valid_action = False
        if verbose > 0: print(f"{state.real_env_steps}: {msg}")

    for unit, transfer_action in actions_by_type["transfer"]:
        valid_action = True
        unit: luxai_unit.Unit
        transfer_action: TransferAction
        if transfer_action.resource > 4 or transfer_action.resource < 0:
            invalidate_action(
                f"Invalid Transfer Action for unit {unit}, transferring invalid resource id {transfer_action.resource}"
            )
            continue
        # if transfer_action.transfer_amount < 0: do not need to check as action space permits range of [0, max_transfer_amount] anyway
        resource_id = transfer_action.resource
        amount = transfer_action.transfer_amount
        if resource_id == 0:
            if unit.cargo.ice < amount:
                invalidate_action(
                    f"Invalid Transfer Action for unit {unit} - Tried to transfer {amount} ice but only had {unit.cargo.ice}"
                )
                continue
        elif resource_id == 1:
            if unit.cargo.ore < amount:
                invalidate_action(
                    f"Invalid Transfer Action for unit {unit} - Tried to transfer {amount} ore but only had {unit.cargo.ore}"
                )
                continue
        elif resource_id == 2:
            if unit.cargo.water < amount:
                invalidate_action(
                    f"Invalid Transfer Action for unit {unit} - Tried to transfer {amount} water but only had {unit.cargo.water}"
                )
                continue
        elif resource_id == 3:
            if unit.cargo.metal < amount:
                invalidate_action(
                    f"Invalid Transfer Action for unit {unit} - Tried to transfer {amount} metal but only had {unit.cargo.metal}"
                )
                continue
        elif resource_id == 4:
            if unit.power < amount:
                invalidate_action(
                    f"Invalid Transfer Action for unit {unit} - Tried to transfer {amount} power but only had {unit.power}"
                )
                continue

        if valid_action:
            actions_by_type_validated["transfer"].append((unit, transfer_action))

    for unit, dig_action in actions_by_type["dig"]:
        valid_action = True
        dig_action: DigAction
        dig_cost = math.ceil(unit.unit_cfg.DIG_COST * weather_cfg["power_loss_factor"])
        if dig_cost > unit.power:
            invalidate_action(
                f"Invalid Dig Action for unit {unit} - Tried to dig requiring ceil({unit.unit_cfg.DIG_COST} x {weather_cfg['power_loss_factor']}) = {dig_cost} power but only had {unit.power} power. Power cost factor is {weather_cfg['power_loss_factor']}"
            )
            continue
        # verify not digging over a factory which is not allowed
        if state.board.factory_occupancy_map[unit.pos.y, unit.pos.x] != -1:
            invalidate_action(
                f"Invalid Dig Action for unit {unit} - Tried to dig on top of a factory"
            )
            continue
        if valid_action:
            actions_by_type_validated["dig"].append((unit, dig_action))

    for unit, pickup_action in actions_by_type["pickup"]:
        valid_action = True
        pickup_action: PickupAction
        unit: luxai_unit.Unit
        factory = state.board.get_factory_at(state, unit.pos)
        if factory is None:
            invalidate_action(f"No factory to pickup from for unit {unit}")
            continue
        resource_id = pickup_action.resource
        amount = pickup_action.pickup_amount
        if resource_id == 0:
            if factory.cargo.ice < amount:
                invalidate_action(
                    f"Invalid Pickup Action for unit {unit} - Tried to pickup {amount} ice but factory only had {factory.cargo.ice}"
                )
                continue
        elif resource_id == 1:
            if factory.cargo.ore < amount:
                invalidate_action(
                    f"Invalid Pickup Action for unit {unit} - Tried to pickup {amount} ore but factory only had {factory.cargo.ore}"
                )
                continue
        elif resource_id == 2:
            if factory.cargo.water < amount:
                invalidate_action(
                    f"Invalid Pickup Action for unit {unit} - Tried to pickup {amount} water but factory only had {factory.cargo.water}"
                )
                continue
        elif resource_id == 3:
            if factory.cargo.metal < amount:
                invalidate_action(
                    f"Invalid Pickup Action for unit {unit} - Tried to pickup {amount} metal but factory only had {factory.cargo.metal}"
                )
                continue
        elif resource_id == 4:
            if factory.power < amount:
                invalidate_action(
                    f"Invalid Pickup Action for unit {unit} - Tried to pickup {amount} power but factory only had {factory.power}"
                )
                continue
        if valid_action:
            actions_by_type_validated["pickup"].append((unit, pickup_action))

    for unit, move_action in actions_by_type["move"]:
        valid_action = True
        move_action: MoveAction
        target_pos: Position = unit.pos + move_action.dist * move_deltas[move_action.move_dir]
        if (
            target_pos.x < 0
            or target_pos.y < 0
            or target_pos.x >= state.board.width
            or target_pos.y >= state.board.height
        ):
            invalidate_action(
                f"Invalid movement action for unit {unit} - Tried to move to {target_pos} which is off the map"
            )
            continue
        if state.board.factory_occupancy_map[target_pos.y, target_pos.x] != -1:
            factory_id = state.board.factory_occupancy_map[target_pos.y, target_pos.x]
            if f"factory_{factory_id}" not in state.factories[unit.team.agent]:
                # if there is a factory but not same team
                invalidate_action(
                    f"Invalid movement action for unit {unit} - Tried to move to {target_pos} which is on an opponent factory"
                )
                continue
        rubble = state.board.rubble[target_pos.y, target_pos.x]
        power_required = math.ceil(unit.move_power_cost(rubble) * weather_cfg["power_loss_factor"])

        if power_required > unit.power:
            invalidate_action(
                f"Invalid movement action for unit {unit} - Tried to move to {target_pos} requiring ceil({unit.move_power_cost(rubble)} x {weather_cfg['power_loss_factor']}) power but only had {unit.power} power. Power cost factor is {weather_cfg['power_loss_factor']}"
            )
            continue
        if valid_action:
            actions_by_type_validated["move"].append((unit, move_action))

    for unit, self_destruct_action in actions_by_type["self_destruct"]:
        valid_action = True
        self_destruct_action: SelfDestructAction
        power_required = math.ceil(unit.unit_cfg.SELF_DESTRUCT_COST * weather_cfg["power_loss_factor"])
        if power_required > unit.power:
            invalidate_action(
                f"Invalid self destruct action for unit {unit} - Tried to self destruct requiring ceil({unit.unit_cfg.SELF_DESTRUCT_COST} x {weather_cfg['power_loss_factor']}) power but only had {unit.power} power. Power cost factor is {weather_cfg['power_loss_factor']}"
            )
        if valid_action:
            actions_by_type_validated["self_destruct"].append((unit, move_action))

    for factory, build_action in actions_by_type["factory_build"]:
        valid_action = True
        build_action: FactoryBuildAction
        factory: 'Factory'

        unit_cfg = env_cfg.ROBOTS[build_action.unit_type.name]
        if factory.cargo.metal < unit_cfg.METAL_COST:
            invalidate_action(f"Invalid factory build action for factory {factory} - Insufficient metal, factory has {factory.cargo.metal}, but requires {unit_cfg.METAL_COST} to build {build_action.unit_type}")
            continue
        if factory.power < math.ceil(unit_cfg.POWER_COST * weather_cfg["power_loss_factor"]):
            invalidate_action(f"Invalid factory build action for factory {factory} - Insufficient power, factory has {factory.power}, but requires ceil({unit_cfg.POWER_COST} x {weather_cfg['power_loss_factor']}) to build {build_action.unit_type.name}. Power cost factor is {weather_cfg['power_loss_factor']}")
            continue
        if valid_action:
            actions_by_type_validated["factory_build"].append((factory, build_action))
        pass

    for factory, water_action in actions_by_type["factory_water"]:
        valid_action = True
        water_cost = factory.water_cost(env_cfg)
        if water_cost > factory.cargo.water:
            invalidate_action(f"Invalid factory water action for factory {factory} - Insufficient water, factory has {factory.cargo.water}, but requires {water_cost} to water lichen")
            continue
        if valid_action:
            actions_by_type_validated["factory_water"].append((factory, water_action))

    return actions_by_type_validated



# Copyright 2022 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import math
import numpy as np
from os import path
from random import choice, randint, randrange, sample, seed, random
from .helpers import board_agent, Board, ShipyardAction
from kaggle_environments import utils
from kaggle_environments.helpers import Point, Direction


def get_col_row(size, pos):
    return pos % size, pos // size


def get_to_pos(size, pos, direction):
    col, row = get_col_row(size, pos)
    if direction == "NORTH":
        return pos - size if pos >= size else size ** 2 - size + col
    elif direction == "SOUTH":
        return col if pos + size >= size ** 2 else pos + size
    elif direction == "EAST":
        return pos + 1 if col < size - 1 else row * size
    elif direction == "WEST":
        return pos - 1 if col > 0 else (row + 1) * size - 1

def check_path(board, start, dirs, dist_a, dist_b, collection_rate):
    kore = 0
    npv = .98
    current = start
    steps = 2 * (dist_a + dist_b + 2)
    for idx, d in enumerate(dirs):
        for _ in range((dist_a if idx % 2 == 0 else dist_b) + 1):
            current = current.translate(d.to_point(), board.configuration.size)
            kore += int((board.cells.get(current).kore or 0) * collection_rate)
    return math.pow(npv, steps) * kore / (2 * (dist_a + dist_b + 2))

def check_location(board, loc, me):
    if board.cells.get(loc).shipyard and board.cells.get(loc).shipyard.player.id == me.id:
        return 0
    kore = 0
    for i in range(-3, 4):
        for j in range(-3, 4):
            pos = loc.translate(Point(i, j), board.configuration.size)
            kore += board.cells.get(pos).kore or 0
    return kore

def get_closest_enemy_shipyard(board, position, me):
    min_dist = 1000000
    enemy_shipyard = None
    for shipyard in board.shipyards.values():
        if shipyard.player_id == me.id:
            continue
        dist = position.distance_to(shipyard.position, board.configuration.size)
        if dist < min_dist:
            min_dist = dist
            enemy_shipyard = shipyard
    return enemy_shipyard
    
def get_shortest_flight_path_between(position_a, position_b, size, trailing_digits=False):
    mag_x = 1 if position_b.x > position_a.x else -1
    abs_x = abs(position_b.x - position_a.x)
    dir_x = mag_x if abs_x < size/2 else -mag_x
    mag_y = 1 if position_b.y > position_a.y else -1
    abs_y = abs(position_b.y - position_a.y)
    dir_y = mag_y if abs_y < size/2 else -mag_y
    flight_path_x = ""
    if abs_x > 0:
        flight_path_x += "E" if dir_x == 1 else "W"
        flight_path_x += str(abs_x - 1) if (abs_x - 1) > 0 else ""
    flight_path_y = ""
    if abs_y > 0:
        flight_path_y += "N" if dir_y == 1 else "S"
        flight_path_y += str(abs_y - 1) if (abs_y - 1) > 0 else ""
    if not len(flight_path_x) == len(flight_path_y):
        if len(flight_path_x) < len(flight_path_y):
            return flight_path_x + (flight_path_y if trailing_digits else flight_path_y[0])
        else:
            return flight_path_y + (flight_path_x if trailing_digits else flight_path_x[0])
    return flight_path_y + (flight_path_x if trailing_digits or not flight_path_x else flight_path_x[0]) if random() < .5 else flight_path_x + (flight_path_y if trailing_digits or not flight_path_y else flight_path_y[0])

@board_agent
def attacker_agent(board):
    me = board.current_player
    remaining_kore = me.kore
    shipyards = me.shipyards
    size = board.configuration.size
    spawn_cost = board.configuration.spawn_cost
    # randomize shipyard order
    shipyards = sample(shipyards, len(shipyards))
    for idx, shipyard in enumerate(shipyards):
        closest_enemy_shipyard = get_closest_enemy_shipyard(board, shipyard.position, me)
        if closest_enemy_shipyard and (remaining_kore >= spawn_cost or shipyard.ship_count >= 50):
            if shipyard.ship_count >= 50:
                flight_plan = get_shortest_flight_path_between(shipyard.position, closest_enemy_shipyard.position, size)
                shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(50, flight_plan)
            elif remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))
        elif shipyard.ship_count >= 21:
            best_h = 0
            best_gap1 = 5
            best_gap2 = 5
            start_dir = board.step % 4
            dirs = Direction.list_directions()[start_dir:] + Direction.list_directions()[:start_dir]
            for gap1 in range(0, 10):
                for gap2 in range(0, 10):
                    h = check_path(board, shipyard.position, dirs, gap1, gap2, .2)
                    if h > best_h:
                        best_h = h
                        best_gap1 = gap1
                        best_gap2 = gap2
            gap1 = str(best_gap1)
            gap2 = str(best_gap2)
            flight_plan = Direction.list_directions()[start_dir].to_char()
            if int(gap1):
                flight_plan += gap1
            next_dir = (start_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            if int(gap2):
                flight_plan += gap2
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            if int(gap1):
                flight_plan += gap1
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(21, flight_plan)
        elif shipyard.ship_count > 0 and len(shipyards) > 1:
            flight_plan = get_shortest_flight_path_between(shipyard.position, shipyards[(idx + 1) % len(shipyards)].position, size)
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(shipyard.ship_count, flight_plan)



@board_agent
def balanced_agent(board):
    me = board.current_player
    me = board.current_player
    remaining_kore = me.kore
    shipyards = me.shipyards
    convert_cost = board.configuration.convert_cost
    size = board.configuration.size
    spawn_cost = board.configuration.spawn_cost

    # randomize shipyard order
    shipyards = sample(shipyards, len(shipyards))
    for shipyard in shipyards:
        closest_enemy_shipyard = get_closest_enemy_shipyard(board, shipyard.position, me)
        invading_fleet_size = 100
        dist_to_closest_enemy_shipyard = 100 if not closest_enemy_shipyard else shipyard.position.distance_to(closest_enemy_shipyard.position, size)
        if closest_enemy_shipyard and (closest_enemy_shipyard.ship_count < 20 or dist_to_closest_enemy_shipyard < 15) and (remaining_kore >= spawn_cost or shipyard.ship_count >= invading_fleet_size):
            if shipyard.ship_count >= invading_fleet_size:
                flight_plan = get_shortest_flight_path_between(shipyard.position, closest_enemy_shipyard.position, size)
                shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(invading_fleet_size, flight_plan)
            elif remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))

        elif remaining_kore > 500 and shipyard.max_spawn > 5:
            if shipyard.ship_count >= convert_cost + 7:
                start_dir = randint(0, 3)
                next_dir = (start_dir + 1) % 4
                best_kore = 0
                best_gap1 = 0
                best_gap2 = 0
                for gap1 in range(5, 15, 3):
                    for gap2 in range(5, 15, 3):
                        gap2 = randint(3, 9)
                        diff1 = Direction.from_index(start_dir).to_point() * gap1
                        diff2 = Direction.from_index(next_dir).to_point() * gap2
                        diff = diff1 + diff2
                        pos = shipyard.position.translate(diff, board.configuration.size)
                        h = check_location(board, pos, me)
                        if h > best_kore:
                            best_kore = h
                            best_gap1 = gap1
                            best_gap2 = gap2
                gap1 = str(best_gap1)
                gap2 = str(best_gap2)
                flight_plan = Direction.list_directions()[start_dir].to_char() + gap1
                flight_plan += Direction.list_directions()[next_dir].to_char() + gap2
                flight_plan += "C"
                shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(max(convert_cost + 7, int(shipyard.ship_count/2)), flight_plan)
            elif remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))

        # launch a large fleet if able
        elif shipyard.ship_count >= 21:
            best_h = 0
            best_gap1 = 5
            best_gap2 = 5
            start_dir = board.step % 4
            dirs = Direction.list_directions()[start_dir:] + Direction.list_directions()[:start_dir]
            for gap1 in range(0, 10):
                for gap2 in range(0, 10):
                    h = check_path(board, shipyard.position, dirs, gap1, gap2, .2)
                    if h > best_h:
                        best_h = h
                        best_gap1 = gap1
                        best_gap2 = gap2
            gap1 = str(best_gap1)
            gap2 = str(best_gap2)
            flight_plan = Direction.list_directions()[start_dir].to_char()
            if int(gap1):
                flight_plan += gap1
            next_dir = (start_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            if int(gap2):
                flight_plan += gap2
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            if int(gap1):
                flight_plan += gap1
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(21, flight_plan)
        # else spawn if possible
        elif remaining_kore > board.configuration.spawn_cost * shipyard.max_spawn:
            remaining_kore -= board.configuration.spawn_cost
            if remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))

@board_agent
def do_nothing_agent(board):
    pass

@board_agent
def random_agent(board):
    me = board.current_player
    remaining_kore = me.kore
    shipyards = me.shipyards
    # randomize shipyard order
    shipyards = sample(shipyards, len(shipyards))
    for shipyard in shipyards:
        # 25% chance to launch a large fleet
        if randint(0, 3) == 0 and shipyard.ship_count > 10:
            dir_str = Direction.random_direction().to_char()
            dir2_str = Direction.random_direction().to_char()
            flight_plan = dir_str + str(randint(1, 10)) + dir2_str
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(min(10, math.floor(shipyard.ship_count / 2)), flight_plan)
        # else spawn if possible
        elif remaining_kore > board.configuration.spawn_cost * shipyard.max_spawn:
            remaining_kore -= board.configuration.spawn_cost
            shipyard.next_action = ShipyardAction.spawn_ships(shipyard.max_spawn)
        # else launch a small fleet
        elif shipyard.ship_count >= 2:
            dir_str = Direction.random_direction().to_char()
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(2, dir_str)

@board_agent
def miner_agent(board):
    me = board.current_player
    remaining_kore = me.kore
    shipyards = me.shipyards
    convert_cost = board.configuration.convert_cost
    spawn_cost = board.configuration.spawn_cost
    # randomize shipyard order
    shipyards = sample(shipyards, len(shipyards))
    for shipyard in shipyards:
        if remaining_kore > 1000 and shipyard.max_spawn > 5:
            if shipyard.ship_count >= convert_cost + 10:
                gap1 = str(randint(3, 9))
                gap2 = str(randint(3, 9))
                start_dir = randint(0, 3)
                flight_plan = Direction.list_directions()[start_dir].to_char() + gap1
                next_dir = (start_dir + 1) % 4
                flight_plan += Direction.list_directions()[next_dir].to_char() + gap2
                next_dir = (next_dir + 1) % 4
                flight_plan += "C"
                shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(max(convert_cost + 10, int(shipyard.ship_count/2)), flight_plan)
            elif remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))

        # launch a large fleet if able

        elif shipyard.ship_count >= 21:
            gap1 = str(randint(3, 9))
            gap2 = str(randint(3, 9))
            start_dir = randint(0, 3)
            flight_plan = Direction.list_directions()[start_dir].to_char() + gap1
            next_dir = (start_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char() + gap2
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char() + gap1
            next_dir = (next_dir + 1) % 4
            flight_plan += Direction.list_directions()[next_dir].to_char()
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(21, flight_plan)
        # else spawn if possible
        elif remaining_kore > board.configuration.spawn_cost * shipyard.max_spawn:
            remaining_kore -= board.configuration.spawn_cost
            if remaining_kore >= spawn_cost:
                shipyard.next_action = ShipyardAction.spawn_ships(min(shipyard.max_spawn, int(remaining_kore/spawn_cost)))
        # else launch a small fleet
        elif shipyard.ship_count >= 2:
            dir_str = Direction.random_direction().to_char()
            shipyard.next_action = ShipyardAction.launch_fleet_with_flight_plan(2, dir_str)

        
agents = {"random": random_agent, "miner": miner_agent, "do_nothing": do_nothing_agent, "balanced": balanced_agent, "attacker": attacker_agent}


def populate_board(state, env):
    obs = state[0].observation
    config = env.configuration
    size = env.configuration.size
    uid_counter = 0

    # Set seed for random number generators
    if not hasattr(config, "randomSeed"):
        max_int_32 = (1 << 31) - 1
        config.randomSeed = randrange(max_int_32)

    np.random.seed(config.randomSeed)
    seed(config.randomSeed)

    # This is a consistent way to generate unique strings to form ship and shipyard ids
    def create_uid():
        nonlocal uid_counter
        uid_counter += 1
        return f"{obs.step}-{uid_counter}"

    # Distribute Kore evenly into quartiles.
    half = math.ceil(size / 2)
    grid = [[0] * half for _ in range(half)]

    # Randomly place a few kore "seeds".
    for i in range(half):
        # random distribution across entire quartile
        grid[randint(0, half - 1)][randint(0, half - 1)] = i ** 2

        # as well as a particular distribution weighted toward the center of the map
        grid[randint(half // 2, half - 1)][randint(half // 2, half - 1)] = i ** 2

    # Spread the seeds radially.
    radius_grid = copy.deepcopy(grid)
    for r in range(half):
        for c in range(half):
            value = grid[r][c]
            if value == 0:
                continue

            # keep initial seed values, but constrain radius of clusters
            radius = min(round((value / half) ** 0.5), 1)
            for r2 in range(r - radius + 1, r + radius):
                for c2 in range(c - radius + 1, c + radius):
                    if 0 <= r2 < half and 0 <= c2 < half:
                        distance = (abs(r2 - r) ** 2 + abs(c2 - c) ** 2) ** 0.5
                        radius_grid[r2][c2] += int(value / max(1, distance) ** distance)

    # add some random sprouts of kore
    radius_grid = np.asarray(radius_grid)
    add_grid = np.random.gumbel(0, 300.0, size=(half, half)).astype(int)
    sparse_radius_grid = np.random.binomial(1, 0.5, size=(half, half))
    add_grid = np.clip(add_grid, 0, a_max=None) * sparse_radius_grid
    radius_grid += add_grid

    # add another set of random locations to the center corner
    corner_grid = np.random.gumbel(0, 500.0, size=(half // 4, half // 4)).astype(int)
    corner_grid = np.clip(corner_grid, 0, a_max=None)
    radius_grid[half - (half // 4):, half - (half // 4):] += corner_grid

    # make it assomptocially symmetric
    for i in range(half):
        for j in range(half):
            if i + j < half:
                radius_grid[i][j] = radius_grid[j][i]

    # Normalize the available kore against the defined configuration starting kore.
    total = sum([sum(row) for row in radius_grid])
    obs.kore = [0] * (size ** 2)
    for r, row in enumerate(radius_grid):
        for c, val in enumerate(row):
            val = int(val * config.startingKore / total / 4)
            obs.kore[size * r + c] = val
            obs.kore[size * r + (size - c - 1)] = val
            obs.kore[size * (size - 1) - (size * r) + c] = val
            obs.kore[size * (size - 1) - (size * r) + (size - c - 1)] = val

    # Distribute the starting shipyards evenly.
    num_agents = len(state)
    starting_positions = [0] * num_agents
    if num_agents == 1:
        starting_positions[0] = size * (size // 2) + size // 2
    elif num_agents == 2:
        starting_positions[0] = size * (size // 2 - size // 4) + size // 4
        starting_positions[1] = size * (size // 2 + size // 4) + math.ceil(3 * size / 4) - 1
    elif num_agents == 4:
        starting_positions[0] = size * (size // 4 + 1) + size // 4 - 1
        starting_positions[1] = size * (size // 4 - 1) + 3 * size // 4 - 1
        starting_positions[2] = size * (3 * size // 4 + 1) + size // 4 + 1
        starting_positions[3] = size * (3 * size // 4 - 1) + 3 * size // 4 + 1
    
    # clear the kore on the starting square
    for pos in starting_positions:
        obs.kore[pos] = 0

    # Initialize the players.
    obs.players = []
    for i in range(num_agents):
        shipyards = {create_uid(): [starting_positions[i], 0, 0]}
        obs.players.append([state[0].reward, shipyards, {}])

    return state


def interpreter(state, env):
    obs = state[0].observation
    config = env.configuration

    # Initialize the board (place cell kore and starting ships).
    if env.done:
        return populate_board(state, env)

    # Interpreter invoked here
    actions = [agent.action for agent in state]
    board = Board(obs, config, actions)
    board = board.next()
    state[0].observation = obs = utils.structify(board.observation)

    # Remove players with invalid status or insufficient potential.
    for index, agent in enumerate(state):
        player_kore, shipyards, fleets = obs.players[index]
        ships_in_shipyards = [int(s[1]) for s in shipyards.values()]
        can_spawn = len(shipyards) > 0 and player_kore >= config.spawnCost
        if agent.status == "ACTIVE" and len(shipyards) == 0 and len(fleets) == 0:
            # Agent can no longer gather any kore
            agent.status = "DONE"
            agent.reward = board.step - board.configuration.episode_steps - 1
        if agent.status == "ACTIVE" and ships_in_shipyards == 0 and len(fleets) == 0 and not can_spawn:
            # Agent can no longer gather any kore
            agent.status = "DONE"
            agent.reward = board.step - board.configuration.episode_steps - 1
        if agent.status != "ACTIVE" and agent.status != "DONE":
            obs.players[index] = [0, {}, {}]

    # Check if done (< 2 players and num_agents > 1)
    if len(state) > 1 and sum(1 for agent in state if agent.status == "ACTIVE") < 2:
        for agent in state:
            if agent.status == "ACTIVE":
                agent.status = "DONE"

    # Update Rewards.
    for index, agent in enumerate(state):
        if agent.status == "ACTIVE":
            agent.reward = obs.players[index][0]
        elif agent.status != "DONE":
            agent.reward = 0

    return state


def renderer(state, env):
    config = env.configuration
    size = config.size
    obs = state[0].observation

    board = [[h, -1, -1, -1] for h in obs.kore]
    for index, player in enumerate(obs.players):
        _, shipyards, fleets = player
        for shipyard in shipyards.values():
            shipyard_pos, _, _ = shipyard
            board[shipyard_pos][1] = index
        for fleet in fleets.values():
            fleet_pos, fleet_kore, ship_count, _, _ = fleet
            board[fleet_pos][2] = index
            board[fleet_pos][3] = ship_count

    col_divider = "|"
    row_divider = "+" + "+".join(["----"] * size) + "+\n"

    out = row_divider
    for row in range(size):
        for col in range(size):
            _, _, fleet, fleet_kore = board[col + row * size]
            out += col_divider + (
                f"{min(int(fleet_kore), 99)}S{fleet}" if fleet > -1 else ""
            ).ljust(4)
        out += col_divider + "\n"
        for col in range(size):
            kore, shipyard, _, _ = board[col + row * size]
            if shipyard > -1:
                out += col_divider + f"SY{shipyard}".ljust(4)
            else:
                out += col_divider + str(min(int(kore), 9999)).rjust(4)
        out += col_divider + "\n" + row_divider

    return out


dir_path = path.dirname(__file__)
json_path = path.abspath(path.join(dir_path, "kore_fleets.json"))
with open(json_path) as json_file:
    specification = json.load(json_file)


def html_renderer():
    js_path = path.abspath(path.join(dir_path, "kore_fleets.js"))
    with open(js_path) as js_file:
        return js_file.read()



import json
import sys
from typing import Dict
if __package__ == "":
    from lux.config import EnvConfig
    from lux.kit import GameState, process_obs, to_json, from_json, process_action, obs_to_game_state
else:
    from .lux.config import EnvConfig
    from .lux.kit import GameState, process_obs, to_json, from_json, process_action, obs_to_game_state
import numpy as np
class Agent():
    def __init__(self, player: str) -> None:
        self.player = player
        self.opp_player = ""
        self.game_state: GameState = None
        if self.player == "player_0":
            self.opp_player = "player_1"
        else:
            self.opp_player = "player_0"
        np.random.seed(0)
        self.step = -1
        self.factories_owned = 0
        self.init_metal_left = 0
        self.init_water_left = 0
        self.env_cfg: EnvConfig = None

    def early_setup(self, step, obs, remainingOverageTime: int):
        """
        Logic here to make actions in the early game. Select faction, bid for an extra factory, and place factories
        """
        # various maps to help aid in decision making over factory placement
        rubble = obs["board"]["rubble"]
        # if ice[y][x] > 0, then there is an ice tile at (x, y)
        ice = obs["board"]["ice"]
        # if ore[y][x] > 0, then there is an ore tile at (x, y)
        ore = obs["board"]["ore"]

        env_cfg = self.env_cfg # the current env configuration for the episode
        
        if step == 0:
            # decide on a faction, and make a bid for the extra factory. 
            # Each unit of bid removes one unit of water and metal from your initial pool
            faction = "MotherMars"
            if self.player == "player_1":
                faction = "AlphaStrike"
            return dict(faction=faction, bid=10)
        else:
            # decide on where to spawn the next factory. Returning an empty dict() will skip your factory placement

            # how much water and metal you have in your starting pool to give to new factories
            water_left = obs["teams"][self.player]["water"]
            metal_left = obs["teams"][self.player]["metal"]
            # how many factories you have left to place
            factories_to_place = obs["teams"][self.player]["factories_to_place"]
            if step == 1:
                # first step of factory placement, we save our initial pool of water and factory amount
                self.factories_owned = factories_to_place
                self.init_metal_left = metal_left
                self.init_water_left = water_left
            # obs["teams"][self.opp_player] has the same information but for the other team
            # potential spawnable locations in your half of the map
            potential_spawns = obs["board"]["spawns"][self.player]

            # as a naive approach we randomly select a spawn location and spawn a factory there
            spawn_loc = potential_spawns[np.random.randint(0, len(potential_spawns))]
            return dict(spawn=spawn_loc, metal=self.init_metal_left // self.factories_owned, water=self.init_water_left // self.factories_owned)

    def act(self, step: int, obs: Dict, remainingOverageTime: int):
        """
        Logic here to make actions for the rest of the game.

        Parameters
        ----------
        step - the current environment step. This is the "real" time step, starts from 0
            after all the bidding and factory placement is done.

        obs - the env observation. This is the raw dictionary based observation and it's usage is detailed below.
            The more raw format is suitable for those who may need it for ML based solutions. For programmed solutions
            an API is provided that interfaces with the observation more nicely, in addition to providing
            functions to generate action vectors

        remainingOverageTime - amount of time in seconds left to make an action in addition to the 
            2s timer per action. Time spent > 2s eats into the remainingOverageTime pool for next turn


        NOTE that step is set back to 0 in this part of the code. This act function is executed max_episode_length times.
        The code ignores the time spent bidding and placing factories
        """
        actions = dict()

        env_cfg: EnvConfig = self.env_cfg # the current env configuration for the episode
        game_state = obs_to_game_state(step=self.step, env_cfg=env_cfg, obs=obs)

        # the weather schedule, a sequence of values representing what the weather is at each real time step
        # 0 = Nothing, 1 = MARS_QUAKE, 2 = COLD_SNAP, 3 = DUST_STORM, 4 = SOLAR_FLARE
        weather_schedule = game_state.weather_schedule
        current_weather = game_state.weather_schedule[step]

        # various maps to help aid in decision making
        rubble = game_state.board.rubble

        # if ice[y][x] > 0, then there is an ice tile at (x, y)
        ice = game_state.board.ice
        # if ore[y][x] > 0, then there is an ore tile at (x, y)
        ore = game_state.board.ore

        # lichen[y][x] = amount of lichen at tile (x, y)
        lichen = game_state.board.lichen
        # lichenStrains[y][x] = the strain id of the lichen at tile (x, y). Each strain id is
        # associated with a single factory and cannot mix with other strains. 
        # factory.strain_id defines the factory's strain id
        lichen_strains = game_state.board.lichen_strains

        # units and factories for your team and the opposition team
        units = game_state.units[self.player]
        opp_units = game_state.units[self.opp_player]
        factories = game_state.factories[self.player]
        opp_factories = game_state.factories[self.opp_player]
        
        # iterate over all active factories
        for unit_id, factory in factories.items():
            if step % 4 == 0 and step > 1:
                p = np.random.uniform()
                if p < 0.75:
                    if factory.can_build_light(game_state):
                        # check if factory can build a light unit, and submit the action if so
                        actions[unit_id] = factory.build_light()
                else:
                    if factory.can_build_heavy(game_state):
                        actions[unit_id] = factory.build_heavy()
            else:
                # find the water cost to grow lichen and only water the lichen if we have enough water
                water_cost = factory.water_cost(game_state)
                if factory.cargo.water > water_cost + 10:
                    actions[unit_id] = factory.water()
                
        for unit_id, unit in units.items():
            # (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)
            move_dir = np.random.randint(0, 5)
            move_cost = unit.move_cost(game_state, move_dir)
            # we attempt to move if move_cost is not None (valid movement) and we have enough power to submit a new action queue and move
            if move_cost is not None and unit.power >= move_cost + unit.action_queue_cost(game_state):
                # by default, any unit action has repeat=True and moves completed actions back to the end of the action queue
                # You can also queue up to env_cfg.UNIT_ACTION_QUEUE_SIZE *actions* for each unit.
                # You can submit action queues for every unit, but be wary that each *action queue* submission costs an additional
                # env_cfg.UNIT_ACTION_QUEUE_POWER_COST[unit.unit_type] power

                # here, we tell this unit to move in a direction once and stay still until controlled again
                actions[unit_id] = [unit.move(move_dir, repeat=False)]
        return actions

### DO NOT REMOVE THE FOLLOWING CODE ###
agent_dict = dict() # store potentially multiple dictionaries as kaggle imports code directly
def agent_fn(observation, configurations):
    """
    agent definition for kaggle submission.
    """
    global agent_dict
    step = observation.step
    
    
    player = observation.player
    remainingOverageTime = observation.remainingOverageTime
    if step == 0:
        agent_dict[player] = Agent(player)
        agent = agent_dict[player]
        # TODO verify this works with kaggle input and add logic to fix that
        agent.env_cfg = EnvConfig.from_dict(configurations["env_cfg"])
    agent = agent_dict[player]
    new_game_state = process_obs(player, agent.game_state, step, json.loads(observation.obs))
    agent.game_state = new_game_state
    agent.step = step
    if step <= agent.game_state["board"]["factories_per_team"] + 1:
        actions = agent.early_setup(step, new_game_state, remainingOverageTime)
    else:
        real_env_steps = new_game_state["real_env_steps"]
        actions = agent.act(real_env_steps, new_game_state, remainingOverageTime)

    return process_action(actions)



from subprocess import Popen, PIPE
from threading  import Thread
from queue import Queue, Empty

import atexit
import os
import sys
agent_processes = [None, None, None, None]
t = None
q = None
def cleanup_process():
    global agent_processes
    for proc in agent_processes:
        if proc is not None:
            proc.kill()
def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        queue.put(line)
    out.close()

def agent(observation, configuration):
    global agent_processes, t, q

    agent_process = agent_processes[observation.player]
    ### Do not edit ###
    if agent_process is None:
        if "__raw_path__" in configuration:
            cwd = os.path.dirname(configuration["__raw_path__"])
        else:
            cwd = os.path.dirname(__file__)
        agent_process = Popen(["node", "dist/MinerBot.js"], stdin=PIPE, stdout=PIPE, stderr=PIPE, cwd=cwd)
        agent_processes[observation.player] = agent_process
        atexit.register(cleanup_process)

        # following 4 lines from https://stackoverflow.com/questions/375427/a-non-blocking-read-on-a-subprocess-pipe-in-python
        q = Queue()
        t = Thread(target=enqueue_output, args=(agent_process.stderr, q))
        t.daemon = True # thread dies with the program
        t.start()
    
    # print observations to agent
    import json
    agent_process.stdin.write((json.dumps(observation) + "\n").encode())
    agent_process.stdin.write((json.dumps(configuration) + "\n").encode())
    agent_process.stdin.flush()

    # wait for data written to stdout
    agent1res = (agent_process.stdout.readline()).decode()

    while True:
        try:  line = q.get_nowait()
        except Empty:
            # no standard error received, break
            break
        else:
            # standard error output received, print it out
            print(line.decode(), file=sys.stderr, end='')

    agent1res = agent1res.strip()
    outputs = agent1res.split(",")
    actions = {}
    for cmd in outputs:
        if cmd != "":
            shipyard_id, action_str = cmd.split(":")
            actions[shipyard_id] = action_str
    return actions



import requests

from typing import *

base_url = "https://www.kaggle.com/requests/EpisodeService/"
get_url = base_url + "GetEpisodeReplay"
list_url = base_url + "ListEpisodes"


def get_episode_replay(episode_id: int):
    body = {
        "EpisodeId": episode_id
    }

    response = requests.post(get_url, json=body)
    return response.json()


def list_episodes(episode_ids: List[int]):
    return __list_episodes({
        "Ids": episode_ids
    })


def list_episodes_for_team(team_id: int):
    return __list_episodes({
        "TeamId": team_id
    })


def list_episodes_for_submission(submission_id: int):
    return __list_episodes({
        "SubmissionId": submission_id
    })


def __list_episodes(body):
    response = requests.post(list_url, json=body)
    return response.json()



import sys
import numpy as np

def get_weather_config(current_weather, env_cfg):
    if type(current_weather) == int or isinstance(current_weather, (int, np.integer)):
        current_weather = env_cfg.WEATHER_ID_TO_NAME[current_weather]
    if current_weather == "MARS_QUAKE":
        return dict(power_gain_factor=1, power_loss_factor=1)
    elif current_weather == "COLD_SNAP":
        return dict(power_gain_factor=1, power_loss_factor=env_cfg.WEATHER["COLD_SNAP"]["POWER_CONSUMPTION"])
    elif current_weather == "DUST_STORM":
        return dict(power_gain_factor=env_cfg.WEATHER["DUST_STORM"]["POWER_GAIN"], power_loss_factor=1)
    elif current_weather == "SOLAR_FLARE":
        return dict(power_gain_factor=env_cfg.WEATHER["SOLAR_FLARE"]["POWER_GAIN"], power_loss_factor=1)
    return dict(power_gain_factor=1, power_loss_factor=1)


import random


def random_agent(observation, configuration):
    return random.randrange(configuration.banditCount - 1)


def round_robin_agent(observation, configuration):
    return observation.step % configuration.banditCount


agents = {
    "random": random_agent,
    "round_robin": round_robin_agent
}


import json
from typing import Dict
import sys
from argparse import Namespace
from agent import agent_fn
if __name__ == "__main__":
    
    def read_input():
        """
        Reads input from stdin
        """
        try:
            return input()
        except EOFError as eof:
            raise SystemExit(eof)
    step = 0
    player_id = 0
    configurations = None
    i = 0
    while True:
        inputs = read_input()
        obs = json.loads(inputs)
        
        observation = Namespace(**dict(step=obs["step"], obs=json.dumps(obs["obs"]), remainingOverageTime=obs["remainingOverageTime"], player=obs["player"], info=obs["info"]))
        if i == 0:
            configurations = obs["info"]["env_cfg"]
        i += 1
        actions = agent_fn(observation, dict(env_cfg=configurations))
        # send actions to engine
        print(json.dumps(actions))


# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import json
import os
import requests
import sys
import traceback
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO
from time import perf_counter
from urllib.parse import urlparse
from .errors import DeadlineExceeded, InvalidArgument
from .utils import read_file, structify


def is_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def get_last_callable(raw, fallback=None, path=None):
    orig_out = sys.stdout
    buffer = StringIO()
    sys.stdout = buffer

    try:
        code_object = compile(raw, path, "exec")
        env = {}

        # append exec_dir so that way python agents can import other files
        exec_dir = os.path.dirname(path)
        sys.path.append(exec_dir)

        exec(code_object, env)
        sys.path.pop()
        sys.stdout = orig_out
        output = buffer.getvalue()
        if output:
            print(output)
        return [v for v in env.values() if callable(v)][-1]
    except Exception as e:
        sys.stdout = orig_out
        output = buffer.getvalue()
        if output:
            print(output)
        if fallback is not None:
            return fallback
        raise InvalidArgument("Invalid raw Python: " + repr(e))


class UrlAgent:
    def __init__(self, raw, environment_name):
        self.raw = raw
        self.environment_name = environment_name

    def __call__(self, observation, configuration):
        data = {
            "action": "act",
            "configuration": configuration,
            "environment": self.environment_name,
            "state": {
                "observation": observation,
            },
        }
        response = requests.post(url=self.raw, data=json.dumps(data))
        response_json = response.json()
        action = response_json["action"]
        if action == "DeadlineExceeded":
            action = DeadlineExceeded()
        elif isinstance(action, str) and action.startswith("BaseException::"):
            # Deserialize the exception message
            parts = action.split("::", 1)
            action = BaseException(parts[1])
        return action


def build_agent(raw, builtin_agents, environment_name):
    """
    Returns the agent and whether the agent is parallelizable.
    """
    if raw in builtin_agents:
        return builtin_agents[raw], False

    # Already callable.
    if callable(raw):
        return raw, False

    # Not a string, static action.
    if not isinstance(raw, str):
        return lambda: raw, False

    # A URL and will be initialized on the calling server.
    if is_url(raw):
        return UrlAgent(raw, environment_name), True

    # A path exists and attempt to grab the source (fallback to the original string).
    raw_agent = raw
    if os.path.exists(raw):
        raw_agent = read_file(raw, raw)
    elif (len(raw) < 100 and ("/" in raw or "\\" in raw)) or raw < 20:
        raise FileNotFoundError("Could not find : " + raw)

    # Attempt to execute the last callable or just return the string.
    agent = None

    def callable_agent(observation, configuration):
        nonlocal agent
        if agent is None:
            agent = get_last_callable(raw_agent, path=raw) or raw_agent
        configuration["__raw_path__"] = raw
        args = [observation, configuration]
        args = args[:agent.__code__.co_argcount]
        return \
            agent(*args) \
            if callable(agent) \
            else agent

    return callable_agent, False

class Agent:
    def __init__(self, raw, environment):
        self.builtin_agents = environment.agents
        self.configuration = environment.configuration
        self.debug = environment.debug
        self.environment_name = environment.name
        self.raw = raw
        self.agent, self.is_parallelizable = build_agent(self.raw, self.builtin_agents, self.environment_name)

    def act(self, observation):
        args = [
            structify(observation),
            structify(self.configuration)
        ]

        if hasattr(self.agent, "__code__"):
            args = args[:self.agent.__code__.co_argcount]

        # Start the timer.

        with StringIO() as out_buffer, StringIO() as err_buffer, redirect_stdout(out_buffer), redirect_stderr(err_buffer):
            try:
                start = perf_counter()
                action = self.agent(*args)
            except Exception as e:
                traceback.print_exc(file=err_buffer)
                action = e

            out = out_buffer.getvalue()
            err = err_buffer.getvalue()
            # Get the maximum log length
            # Allow up to 1k (default) log characters per step which is ~1MB per 600 step episode
            max_log_length = self.configuration.get('maxLogLength', 1024)

            # truncate if max_log_length is set to None, do not truncate
            if max_log_length is not None:
                out = out[0:max_log_length]
                err = err[0:max_log_length]

        duration = perf_counter() - start
        log = {
            "duration": round(duration, 6),
            "stdout": out,
            "stderr": err,
        }

        if self.debug:
            if not log["stdout"].isspace():
                print(log["stdout"], end="")
            if not log["stderr"].isspace():
                print(log["stderr"], end="")

        if duration - self.configuration.actTimeout > observation.remainingOverageTime:
            # No overage time left, timeout agent
            action = DeadlineExceeded()

        return action, log



import operator
import math
from enum import Enum, auto
import random
from typing import *


class Point(tuple):
    """
    This type wraps Tuple[int, int] to provide additional operators and convenience members.
    Point are expressed in the form (x, y) where x is the board column and y is the row.
    (0, 0) is the lower left corner of the board and (size - 1, size - 1) is the upper right corner of the board.
    Note that this differs from arrays where the top left is (0, 0) and the bottom right is (size - 1, size - 1).
    Note that operators in this class do not constrain points to the board.
    You can generally constrain a point to the board by calling point % board.configuration.size.
    """
    def __new__(cls: Type['Point'], x: int, y: int):
        return super(Point, cls).__new__(cls, tuple((x, y)))

    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    def map(self, f: Callable[[int], int]) -> 'Point':
        return Point(f(self[0]), f(self[1]))

    def map2(self, other: Union[Tuple[int, int], 'Point'], f: Callable[[int, int], int]) -> 'Point':
        return Point(f(self[0], other[0]), f(self[1], other[1]))

    def translate(self, offset: 'Point', size: int):
        """Translates the current point by offset and wraps it around a board of width and height size"""
        return (self + offset) % size

    def distance_to(self, other: 'Point', size: int):
        """Translates the current point by offset and wraps it around a board of width and height size"""
        abs_x = abs(self.x - other.x)
        dist_x = abs_x if abs_x < size/2 else size - abs_x
        abs_y = abs(self.y - other.y)
        dist_y = abs_y if abs_y < size/2 else size - abs_y
        return dist_x + dist_y

    def to_index(self, size: int):
        """
        Converts a 2d position in the form (x, y) to an index in the observation.halite list.
        See index_to_position for the inverse.
        """
        return (size - self.y - 1) * size + self.x

    @staticmethod
    def from_index(index: int, size: int) -> 'Point':
        """
        Converts an index in the observation.halite list to a 2d position in the form (x, y).
        See position_to_index for the inverse.
        """
        y, x = divmod(index, size)
        return Point(x, (size - y - 1))

    def __abs__(self) -> 'Point':
        return self.map(operator.abs)

    def __add__(self, other: Union[Tuple[int, int], 'Point']) -> 'Point':
        return self.map2(other, operator.add)

    def __eq__(self, other: Union[Tuple[int, int], 'Point']) -> bool:
        try:
            return self[0] == other[0] and self[1] == other[1]
        except (TypeError, IndexError):
            return False

    def __floordiv__(self, denominator: int) -> 'Point':
        return self.map(lambda x: x // denominator)

    def __hash__(self) -> int:
        return hash((self.x, self.y))

    def __mod__(self, mod: int) -> 'Point':
        return self.map(lambda x: x % mod)

    def __mul__(self, factor: int) -> 'Point':
        return self.map(lambda x: x * factor)

    def __neg__(self) -> 'Point':
        return self.map(operator.neg)

    def __str__(self):
        return f"({self.x}, {self.y})"

    def __sub__(self, other: Union[Tuple[int, int], 'Point']) -> 'Point':
        return self.map2(other, operator.sub)


class Direction(Enum):
    NORTH = auto()
    EAST = auto()
    SOUTH = auto()
    WEST = auto()

    def to_point(self) -> Point:
        """
        This returns the position offset associated with a particular action 
        NORTH -> (0, 1)
        EAST -> (1, 0)
        SOUTH -> (0, -1)
        WEST -> (-1, 0)
        """
        return (
            Point(0, 1) if self == Direction.NORTH else
            Point(1, 0) if self == Direction.EAST else
            Point(0, -1) if self == Direction.SOUTH else
            Point(-1, 0) if self == Direction.WEST else
            None
        )

    def __str__(self) -> str:
        return self.name

    def to_index(self) -> int:
        return (
            0 if self == Direction.NORTH else
            1 if self == Direction.EAST else
            2 if self == Direction.SOUTH else
            3 if self == Direction.WEST else
            None
        )

    def to_char(self) -> str:
        return (
            "N" if self == Direction.NORTH else
            "E" if self == Direction.EAST else
            "S" if self == Direction.SOUTH else
            "W" if self == Direction.WEST else
            None
        )

    def opposite(self) -> 'Direction':
        return (
            Direction.SOUTH if self == Direction.NORTH else
            Direction.WEST if self == Direction.EAST else
            Direction.NORTH if self == Direction.SOUTH else
            Direction.EAST if self == Direction.WEST else
            None
        )

    def rotate_left(self) -> 'Direction':
        return (
            Direction.WEST if self == Direction.NORTH else
            Direction.NORTH if self == Direction.EAST else
            Direction.EAST if self == Direction.SOUTH else
            Direction.SOUTH if self == Direction.WEST else
            None
        )

    def rotate_right(self) -> 'Direction':
        return (
            Direction.EAST if self == Direction.NORTH else
            Direction.SOUTH if self == Direction.EAST else
            Direction.WEST if self == Direction.SOUTH else
            Direction.NORTH if self == Direction.WEST else
            None
        )

    @staticmethod
    def from_str(str_dir: str):
        return  (
            Direction.NORTH if str_dir == "NORTH" else
            Direction.EAST if str_dir == "EAST" else
            Direction.SOUTH if str_dir == "SOUTH" else
            Direction.WEST if str_dir == "WEST" else
            None
        )

    @staticmethod
    def from_char(str_char: str):
        return  (
            Direction.NORTH if str_char == "N" else
            Direction.EAST if str_char == "E" else
            Direction.SOUTH if str_char == "S" else
            Direction.WEST if str_char == "W" else
            None
        )

    @staticmethod
    def from_index(idx: int):
        return (
            Direction.NORTH if idx == 0 else
            Direction.EAST if idx == 1 else
            Direction.SOUTH if idx == 2 else
            Direction.WEST if idx == 3 else
            None
        )

    @staticmethod
    def random_direction() -> 'Direction':
        rand = random.random()
        if rand <= .25:
            return Direction.NORTH
        elif rand <= .5:
            return Direction.EAST
        elif rand <= .75:
            return Direction.SOUTH
        else:
            return Direction.WEST

    @staticmethod
    def list_directions() -> List['Direction']:
        return [
            Direction.NORTH,
            Direction.EAST,
            Direction.SOUTH,
            Direction.WEST,
        ]


TItem = TypeVar('TItem')
THash = TypeVar('THash')


def group_by(items: Iterable[TItem], selector: Callable[[TItem], THash]) -> Dict[THash, List[TItem]]:
    results = {}
    for item in items:
        key = selector(item)
        if key not in results:
            results[key] = []
        results[key].append(item)
    return results


def histogram(items: Iterable[TItem]) -> Dict[TItem, int]:
    """Accepts a list of hashable items and returns a dictionary where the keys are items and the values are counts of each item in the list."""
    results = {}
    for item in items:
        if item not in results:
            results[item] = 1
        else:
            results[item] += 1
    return results


def with_print(item: TItem) -> TItem:
    """Prints an item and returns it -- useful for debug printing in lambdas and chained functions."""
    print(item)
    return item


class Observation(Dict[str, any]):
    """
    Observation provides access to per-step parameters in the environment.
    """
    @property
    def step(self) -> int:
        """Current step within the episode."""
        return self["step"]

    @property
    def remaining_overage_time(self) -> int:
        """Total remaining banked time (seconds) that can be used in excess of per-step actTimeouts -- agent is disqualified with TIMEOUT status when this drops below 0."""
        return self["remainingOverageTime"]


class Configuration(Dict[str, any]):
    """
    Configuration provides access to tunable parameters in the environment.
    """
    @property
    def episode_steps(self) -> int:
        """Total number of steps/turns in the run."""
        return self["episodeSteps"]

    @property
    def act_timeout(self) -> float:
        """Maximum runtime (seconds) to obtain an action from an agent."""
        return self["actTimeout"]

    @property
    def run_timeout(self) -> float:
        """Maximum runtime (seconds) of an episode (not necessarily DONE)."""
        return self["runTimeout"]


TConfiguration = TypeVar('TConfiguration', bound=Configuration)
TObservation = TypeVar('TObservation', bound=Observation)
TAction = TypeVar('TAction')
Agent = Callable[[TObservation, TConfiguration], TAction]


class AgentStatus(Enum):
    UNKNOWN = 0
    ACTIVE = 1
    INACTIVE = 2
    DONE = 3
    INVALID = 4
    ERROR = 5


class AgentState(Generic[TObservation, TAction], Dict[str, any]):
    @property
    def observation(self) -> TObservation:
        return self["observation"]

    @property
    def action(self) -> TAction:
        return self["action"]

    @property
    def reward(self) -> int:
        return self["reward"]

    @property
    def status(self) -> AgentStatus:
        status = self["status"]
        if status in AgentStatus.__members__:
            return AgentStatus[status]
        return AgentStatus.UNKNOWN


class Environment(Generic[TConfiguration, TObservation, TAction]):
    @property
    def specification(self) -> Dict[str, any]:
        raise NotImplemented()

    def interpret(self, configuration: TConfiguration, state: List[AgentState[TObservation, TAction]]) -> List[AgentState[TObservation, TAction]]:
        raise NotImplemented()

    def render_html(self, configuration: TConfiguration, state: List[AgentState[TObservation, TAction]]) -> str:
        raise NotImplemented()

    def render_text(self, configuration: TConfiguration, state: List[AgentState[TObservation, TAction]]) -> str:
        raise NotImplemented()

    @property
    def builtin_agents(self) -> Dict[str, Agent]:
        """Override this property to provide default agents that can be referenced by name in this environment, e.g. `{"random": my_random_agent}`"""
        return {}



from __future__ import annotations
from collections import defaultdict
from typing import Dict, List
import numpy as np
from typing import TYPE_CHECKING
from luxai2022.config import EnvConfig
if TYPE_CHECKING:
    from luxai2022.factory import Factory
    from luxai2022.state import State
from luxai2022.map.position import Position
from luxai2022.map_generator.generator import GameMap
from luxai2022.unit import Unit



class Board:
    def __init__(self, seed=None, env_cfg: EnvConfig = None) -> None:
        self.env_cfg = env_cfg
        self.height =  self.env_cfg.map_size
        self.width =  self.env_cfg.map_size
        self.seed = seed
        rng = np.random.RandomState(seed=seed)
        map_type = rng.choice(["Cave", "Mountain"])
        symmetry = rng.choice(["horizontal", "vertical"])
        self.map = GameMap.random_map(seed=seed, symmetry=symmetry, map_type=map_type, width=self.width, height=self.height)
        self.factories_per_team = rng.randint(env_cfg.MIN_FACTORIES, env_cfg.MAX_FACTORIES + 1)

        # remove bottom once generator is fully ready
        self.map.rubble = self.map.rubble.astype(int)
        self.map.ore = self.map.ore.astype(int)
        self.map.ice = self.map.ice.astype(int)

        self.lichen = np.zeros((self.height, self.width), dtype=int)
        # ownership of lichen by factory id, a simple mask
        # -1 = no ownership
        self.lichen_strains = -np.ones((self.height, self.width), dtype=int)
        # self.units_map: np.ndarray = np.zeros((self.height, self.width))
        self.units_map: Dict[str, List[Unit]] = defaultdict(list)
        # TODO consider unit occupancy map, may speed up computation by storing more info in a form indexable in numpy arrays

        # maps center of factory to the factory
        self.factory_map: Dict[str, 'Factory'] = dict()
        # != -1 if a factory tile is on the location. Equals factory number id
        self.factory_occupancy_map = -np.ones((self.height, self.width), dtype=int)


        # Valid spawn locations
        player_0_spawn_info = self.get_valid_spawns(0)
        player_1_spawn_info = self.get_valid_spawns(1)
        self.spawns = {"player_0": player_0_spawn_info[0],
                       "player_1": player_1_spawn_info[0]}
        self.spawn_masks = {"player_0": player_0_spawn_info[1],
                       "player_1": player_1_spawn_info[1]}
    def pos_hash(self, pos: Position):
        return f"{pos.x},{pos.y}"
    def get_units_at(self, pos: Position):
        pos_hash = self.pos_hash(pos)
        if pos_hash in self.units_map:
            return self.units_map[pos_hash]
        return None
    def get_factory_at(self, state: State, pos: Position):
        f_id = self.factory_occupancy_map[pos.y, pos.x]
        if f_id != -1:
            unit_id = f"factory_{f_id}"
            if unit_id in state.factories["player_0"]:
                return state.factories["player_0"][unit_id]
            else:
                return state.factories["player_1"][unit_id]
        return None
    def get_valid_spawns(self, team_id):
        xx, yy = np.mgrid[:self.width, :self.height]
        if self.map.symmetry == "vertical":
            if team_id == 0:
                spawns_mask = xx < (self.width - 2) / 2
            else:
                spawns_mask = xx > (self.width + 2) / 2
        if self.map.symmetry == "horizontal":
            if team_id == 0:
                spawns_mask = yy < (self.height - 2) / 2

            else:
                spawns_mask = yy > (self.height + 2) / 2

        # if self.map.symmetry == "rotational":
        #     if team_id == 0:
        #         spawns_mask = xx < (self.width - 1) / 2

        #     else:
        #         spawns_mask = xx > (self.width - 1) / 2

        # if self.map.symmetry == "/":
        #     if team_id == 0:
        #         spawns_mask = xx - yy < 0

        #     else:
        #         spawns_mask = xx - yy > 0

        # if self.map.symmetry == "\\":
        #     if team_id == 0:
        #         spawns_mask = xx + yy < (self.width + self.height) / 2 - 1

        #     else:
        #         spawns_mask = xx + yy > (self.width + self.height) / 2 - 1

        x, y = np.where(spawns_mask)

        spawns = np.array([*zip(x, y)])
        spawns = spawns[(spawns[:, 0] > 0) & (spawns[:, 1] > 0)]
        spawns = spawns[(spawns[:, 0] < self.width - 1) & (spawns[:, 1] < self.height - 1)]
        return spawns, spawns_mask

    @property
    def rubble(self) -> np.ndarray:
        return self.map.rubble
    @property
    def ice(self) -> np.ndarray:
        return self.map.ice
    @property
    def ore(self):
        return self.map.ore
    def state_dict(self):
        return dict(
            rubble=self.rubble.copy(),
            ore=self.ore.copy(),
            ice=self.ice.copy(),
            lichen=self.lichen.copy(),
            lichen_strains=self.lichen_strains.copy(),
            spawns=self.spawns.copy(),
            factories_per_team=self.factories_per_team,
        )



import math
from sys import stderr
import numpy as np
from dataclasses import dataclass
if __package__ == "":
    from lux.weather import get_weather_config
    from lux.cargo import UnitCargo
    from lux.config import EnvConfig
else:
    from .weather import get_weather_config
    from .cargo import UnitCargo
    from .config import EnvConfig
@dataclass
class Factory:
    team_id: int
    unit_id: str
    strain_id: int
    power: int
    cargo: UnitCargo
    pos: np.ndarray
    # lichen tiles connected to this factory
    # lichen_tiles: np.ndarray
    env_cfg: EnvConfig


    def build_heavy_metal_cost(self, game_state):
        unit_cfg = self.env_cfg.ROBOTS["HEAVY"]
        return unit_cfg.METAL_COST
    def build_heavy_power_cost(self, game_state):
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        unit_cfg = self.env_cfg.ROBOTS["HEAVY"]
        return math.ceil(unit_cfg.POWER_COST * weather_cfg["power_loss_factor"])
    def can_build_heavy(self, game_state):
        return self.power >= self.build_heavy_power_cost(game_state) and self.cargo.metal >= self.build_heavy_metal_cost(game_state)
    def build_heavy(self):
        return 1

    def build_light_metal_cost(self, game_state):
        unit_cfg = self.env_cfg.ROBOTS["LIGHT"]
        return unit_cfg.METAL_COST
    def build_light_power_cost(self, game_state):
        current_weather = game_state.weather_schedule[game_state.real_env_steps]
        weather_cfg = get_weather_config(current_weather, self.env_cfg)
        unit_cfg = self.env_cfg.ROBOTS["LIGHT"]
        return math.ceil(unit_cfg.POWER_COST * weather_cfg["power_loss_factor"])
    def can_build_light(self, game_state):
        return self.power >= self.build_light_power_cost(game_state) and self.cargo.metal >= self.build_light_metal_cost(game_state)

    def build_light(self):
        return 0

    def water_cost(self, game_state):
        """
        Water required to perform water action
        """
        owned_lichen_tiles = (game_state.board.lichen_strains == self.strain_id).sum()
        return np.ceil(owned_lichen_tiles / self.env_cfg.LICHEN_WATERING_COST_FACTOR) + 1
    def can_water(self, game_state):
        return self.cargo.water >= self.water_cost(game_state)
    def water(self):
        return 2



from typing import Any, Dict, List
from gym import spaces
import gym

from luxai2022.config import EnvConfig
from luxai2022.factory import Factory
from luxai2022.team import FactionTypes
from luxai2022.unit import Unit
import random
import numpy as np

class FactionString(gym.Space):
    def __init__(
                self,
            ):
        self.valid_factions = ["Null"] + [x.name for x in FactionTypes]

    def sample(self):
        return random.choice(self.valid_factions)

    def contains(self, x):
        return type(x) == "str" and x in self.valid_factions

class PartialDict(spaces.Dict):
    def contains(self, x: Any) -> bool:
        if not isinstance(x, dict) or len(x) > len(self.spaces):
            return False, "invalid"
        for k, v in x.items():
            if k in self.spaces:
                space = self.spaces[k]
                if not space.contains(v):
                    return False, f"invalid action {v} for {k}"
            else:
                return False, f"{k} is not on your team or does not exist"
        return True, None

class ActionsQueue(spaces.Space):
    def __init__(self, action_space: spaces.Space, max_length: int) -> None:
        super().__init__((), float)
        self.max_length = max_length
        self.action_space = action_space
    def sample(self):
        queue_size = self.np_random.randint(0, self.max_length) + 1
        action_q = []
        for _ in range(queue_size):
            action_q.append(self.action_space.sample())
        return action_q
    def contains(self, x: Any) -> bool:
        if isinstance(x, list) and len(x) > self.max_length:
            return False
        elif isinstance(x, np.ndarray) and len(x.shape) == 2 and len(x) > self.max_length:
            return False
        elif isinstance(x, np.ndarray) and len(x.shape) == 1:
            x = [x]
        # if (not isinstance(x, list) and not isinstance(x, np.ndarray)) or len(x) > self.max_length:
            # return False
        for a in x:
            # fix issue where multidiscrete space does not do 100% check on type of value
            try:
                if not self.action_space.contains(a):
                    return False
            except:
                return False
        return True

def get_act_space_init(config: EnvConfig, agent: str):
    # Get action space for turn 0 initialization
    act_space = dict()
    act_space["faction"] = FactionString()
    act_space["spawns"] = spaces.Box(low=0, high=config.map_size, shape=(config.MAX_FACTORIES, 2), dtype=int)
    return spaces.Dict(act_space)

def get_act_space_bid(config: EnvConfig, agent: str):
    act_space = dict()
    act_space["faction"] = FactionString()
    act_space["bid"] = spaces.Discrete(100000)
    return spaces.Dict(act_space)
def get_act_space_placement(config: EnvConfig, agent: str):
    # Get action space for turn 0 initialization
    act_space = dict()
    act_space["spawn"] = spaces.Box(low=0, high=config.map_size, shape=(2,), dtype=int)
    act_space["water"] = spaces.Discrete(100000)
    act_space["metal"] = spaces.Discrete(100000)
    return spaces.Dict(act_space)

def get_act_space(units: Dict[str, Dict[str, Unit]], factories: Dict[str, Dict[str, Factory]], config: EnvConfig, agent: str):
    act_space = dict()

    # for consistency, every action space per unit is fixed, makes it easier to work out of the box.
    # we further annotate dimensions, in many places for clarity

    # for those who are programming rule-based bots, we provide helper functions to generate the action vector to store in an actions dict, as well as generate
    # a human readable version of an action vector

    # TODO - verify speed of building action spaces like this.
    for u in units[agent].values():
        # Each action for any mobile unit, (light, heavy) is an array A of shape (max_queue, 5)
        # if max_queue is 1, we remove this dimension
        # let a be some element in A
        # Then

        # a[0] = action type
        # (0 = move, 1 = transfer X amount of R, 2 = pickup X amount of R, 3 = dig, 4 = self destruct, 5 = recharge X)

        # a[1] = direction (0 = center, 1 = up, 2 = right, 3 = down, 4 = left)

        # a[2] = R = resource type (0 = ice, 1 = ore, 2 = water, 3 = metal, 4 power)

        # a[3] = X, amount of resources transferred or picked up if action is transfer or pickup.
        # If action is recharge, it is how much energy to store before executing the next action in queue

        # a[4] = 0,1 - repeat false or repeat true. If true, action is sent to end of queue once consumed
        act_space[u.unit_id] = ActionsQueue(spaces.MultiDiscrete([6, 5, 5, config.max_transfer_amount, 2]), config.UNIT_ACTION_QUEUE_SIZE)

    for factory in factories[agent].values():
        # action type (0 = build light robot, 1 = build heavy robot, 2 = water lichen)
        act_space[factory.unit_id] = spaces.Discrete(3)


    return PartialDict(act_space)



from dataclasses import dataclass
from enum import Enum
from luxai2022.config import EnvConfig
from luxai2022.globals import TERM_COLORS
try:
    from termcolor import colored
except:
    pass
@dataclass
class FactionInfo:
    color: str = "none"
    alt_color: str = "red"
    faction_id: int = -1


class FactionTypes(Enum):
    AlphaStrike = FactionInfo(color="yellow", faction_id=0)
    MotherMars = FactionInfo(color="green", faction_id=1)
    TheBuilders = FactionInfo(color="blue", faction_id=2)
    FirstMars = FactionInfo(color="red", faction_id=3)


class Team:
    def __init__(self, team_id: int, agent: str, faction: FactionTypes = None) -> None:
        self.faction = faction
        self.team_id = team_id
        # the key used to differentiate ownership of things in state
        self.agent = agent

        self.init_water = 0
        self.init_metal = 0
        self.factories_to_place = 0
        self.factory_strains = []
    def state_dict(self):
        return dict(
            team_id=self.team_id,
            faction=self.faction.name,
            # note for optimization, water,metal, factories_to_place doesn't change after the early game.
            water=self.init_water,
            metal=self.init_metal,
            factories_to_place=self.factories_to_place,
            factory_strains=self.factory_strains
        )
    def __str__(self) -> str:
        out = f"[Player {self.team_id}]"
        if TERM_COLORS:
            return colored(out, self.faction.value.color)
        return out
